from typing import Generator, Optional, Callable, List, Tuple, Iterator
from collections import defaultdict
from enum import Enum

from astroid import nodes

from edulint.linting.analyses.cfg.graph import CFGBlock, CFGLoc


def unwrap(loc: CFGLoc) -> CFGLoc:
    """unwraps nodes that are not part of the CFG (for, if, while, ...)"""
    return loc.block.locs[loc.pos]


class Direction(Enum):
    SUCCESSORS = 1
    PREDECESSORS = -1

    @staticmethod
    def to_index(pos: int, direction: "Direction") -> int:
        if direction == Direction.SUCCESSORS:
            return pos
        if direction == Direction.PREDECESSORS:
            return -pos - 1
        assert False, "unreachable"

    @staticmethod
    def to_range(i: int, block_len: int, direction: "Direction") -> Iterator[int]:
        if direction == Direction.SUCCESSORS:
            return range(i, block_len)
        if direction == Direction.PREDECESSORS:
            return range(i, -1, -1)
        assert False, "unreachable"

    @staticmethod
    def get_essors(block: CFGBlock, direction: "Direction") -> Generator[CFGBlock, None, None]:
        attr = direction.name.lower()
        for edge in getattr(block, attr):
            yield getattr(edge, "target" if direction == Direction.SUCCESSORS else "source")


def essor_blocks_from_locs(
    locs: List[CFGLoc],
    direction: Direction,
    stop_on_loc: Optional[Callable[[CFGLoc], bool]],
    stop_on_block: Optional[Callable[[CFGBlock, int, int], bool]],
    include_start: bool,
    include_end: bool,
) -> Iterator[Tuple[CFGBlock, int, int]]:

    def triage_block(
        current_block: CFGBlock, from_pos: int, to_pos: int, is_first: bool
    ) -> Tuple[bool, bool, Tuple[int, int]]:
        stop_on_current_block = stop_on_block is not None and stop_on_block(
            current_block, from_pos, to_pos
        )
        if (
            stop_on_current_block
            and not include_end
            and (not include_start or not is_first)
            and stop_on_loc is None
        ):
            return True, False, (None, None)

        if (stop_on_block is None or stop_on_current_block) and stop_on_loc is not None:
            if direction == Direction.SUCCESSORS:
                rng = range(from_pos, to_pos)
            else:
                rng = range(to_pos - 1, from_pos - 1, -1)

            for i in rng:
                current_loc = current_block.locs[i]
                if stop_on_loc is not None and stop_on_loc(current_loc):
                    if include_end:
                        if direction == Direction.SUCCESSORS:
                            to_pos = i + 1
                        else:
                            from_pos = i
                    else:
                        if direction == Direction.SUCCESSORS:
                            to_pos = i
                        else:
                            from_pos = i + 1
                    return True, True, (from_pos, to_pos)

        return stop_on_current_block, True, (from_pos, to_pos)

    def dfs_iter(current_block: CFGBlock, from_pos: int, to_pos: int, is_first: bool):
        stack = [(current_block, from_pos, to_pos, is_first, 0)]

        while stack:
            current_block, from_pos, to_pos, is_first, essor_i = stack.pop()

            if essor_i == 0:
                should_stop, should_yield, (from_pos, to_pos) = triage_block(
                    current_block, from_pos, to_pos, is_first
                )
                if should_yield and from_pos != to_pos:
                    yield current_block, from_pos, to_pos
                if should_stop:
                    continue

            if (direction == Direction.SUCCESSORS and essor_i >= len(current_block.successors)) or (
                direction == Direction.PREDECESSORS and essor_i >= len(current_block.predecessors)
            ):
                continue
            stack.append((current_block, from_pos, to_pos, is_first, essor_i + 1))

            if (direction == Direction.SUCCESSORS and to_pos == len(current_block.locs)) or (
                direction == Direction.PREDECESSORS and from_pos == 0
            ):
                essor = list(Direction.get_essors(current_block, direction))[essor_i]
                if len(essor.locs) == 0:
                    continue

                elif essor not in visited:
                    visited[essor] = 0, len(essor.locs)
                    stack.append((essor, 0, len(essor.locs), False, 0))

                else:
                    vfrom, vto = visited[essor]
                    if vfrom == 0 and vto == len(essor.locs):
                        continue

                    if direction == Direction.SUCCESSORS:
                        assert vto == len(essor.locs)
                        new_from_pos = 0
                        new_to_pos = vfrom
                    else:
                        assert vfrom == 0
                        new_from_pos = vto
                        new_to_pos = len(essor.locs)

                    visited[essor] = 0, len(essor.locs)
                    stack.append((essor, new_from_pos, new_to_pos, False, 0))

    visited = {}
    for loc in locs:
        loc = unwrap(loc)
        vblock = visited.get(loc.block)

        if vblock is None:
            if direction == Direction.SUCCESSORS:
                from_pos = loc.pos if include_start else loc.pos + 1
                to_pos = len(loc.block.locs)
            else:
                from_pos = 0
                to_pos = loc.pos + 1 if include_start else loc.pos

            visited[loc.block] = from_pos, to_pos
            yield from dfs_iter(loc.block, from_pos, to_pos, is_first=True)
        else:
            vfrom, vto = vblock
            if vfrom <= loc.pos < vto:
                continue

            if direction == Direction.SUCCESSORS:
                assert vto == len(loc.block.locs)
                from_pos = loc.pos if include_start else loc.pos + 1
                to_pos = vfrom
                visited[loc.block] = from_pos, vto
            else:
                assert vfrom == 0
                from_pos = vto
                to_pos = loc.pos + 1 if include_start else loc.pos
                visited[loc.block] = vfrom, to_pos

            yield from dfs_iter(loc.block, from_pos, to_pos, is_first=True)


def successor_blocks_from_locs(
    locs: List[CFGLoc],
    stop_on_loc: Optional[Callable[[CFGLoc], bool]] = None,
    stop_on_block: Optional[Callable[[CFGBlock, int, int], bool]] = None,
    include_start: bool = False,
    include_end: bool = False,
):
    return essor_blocks_from_locs(
        locs, Direction.SUCCESSORS, stop_on_loc, stop_on_block, include_start, include_end
    )


def predecessor_blocks_from_locs(
    locs: List[CFGLoc],
    stop_on_loc: Optional[Callable[[CFGLoc], bool]] = None,
    stop_on_block: Optional[Callable[[CFGBlock, int, int], bool]] = None,
    include_start: bool = False,
    include_end: bool = False,
):
    return essor_blocks_from_locs(
        locs, Direction.PREDECESSORS, stop_on_loc, stop_on_block, include_start, include_end
    )


def essor_locs_from_locs(
    locs: List[CFGLoc],
    direction: Direction,
    stop_on_loc: Optional[Callable[[CFGLoc], bool]],
    stop_on_block: Optional[Callable[[CFGBlock], bool]],
    include_start: bool,
    include_end: bool,
    explore_functions: bool,
    explore_classes: bool,
):
    def explore_function(node: nodes.FunctionDef):
        assert direction == Direction.SUCCESSORS
        yield from essor_locs_from_locs(
            [node.args.cfg_loc],
            direction,
            stop_on_loc,
            stop_on_block,
            include_start,
            include_end,
            explore_functions,
            explore_classes,
        )

    def try_explore_function(loc: CFGLoc):
        if explore_functions and isinstance(loc.node, nodes.FunctionDef):
            yield from explore_function(loc.node)

    def explore_class(node: nodes.ClassDef):
        assert direction == Direction.SUCCESSORS
        for child in node.body:
            yield from try_explore_function(child.cfg_loc)

    def try_explore_class(loc: CFGLoc):
        if explore_classes and isinstance(loc.node, nodes.ClassDef):
            yield from explore_class(loc.node)

    for block, from_pos, to_pos in essor_blocks_from_locs(
        locs, direction, stop_on_loc, stop_on_block, include_start, include_end
    ):
        if direction == Direction.SUCCESSORS:
            rng = range(from_pos, to_pos)
        else:
            rng = range(to_pos - 1, from_pos - 1, -1)
        for i in rng:
            yield block.locs[i]
            yield from try_explore_function(block.locs[i])
            yield from try_explore_class(block.locs[i])


def successors_from_loc(
    loc: CFGLoc,
    stop_on_loc: Optional[Callable[[CFGLoc], bool]] = None,
    stop_on_block: Optional[Callable[[CFGBlock], bool]] = None,
    include_start: bool = False,
    include_end: bool = False,
    explore_functions: bool = False,
    explore_classes: bool = False,
) -> Generator[CFGLoc, None, None]:
    yield from essor_locs_from_locs(
        [loc],
        Direction.SUCCESSORS,
        stop_on_loc,
        stop_on_block,
        include_start,
        include_end,
        explore_functions,
        explore_classes,
    )


def successors_from_locs(
    locs: List[CFGLoc],
    stop_on_loc: Optional[Callable[[CFGLoc], bool]] = None,
    stop_on_block: Optional[Callable[[CFGBlock], bool]] = None,
    include_start: bool = False,
    include_end: bool = False,
    explore_functions: bool = False,
    explore_classes: bool = False,
) -> Generator[CFGLoc, None, None]:
    yield from essor_locs_from_locs(
        locs,
        Direction.SUCCESSORS,
        stop_on_loc,
        stop_on_block,
        include_start,
        include_end,
        explore_functions,
        explore_classes,
    )


def predecessors_from_loc(
    loc: CFGLoc,
    stop_on_loc: Optional[Callable[[CFGLoc], bool]] = None,
    stop_on_block: Optional[Callable[[CFGBlock], bool]] = None,
    include_start: bool = False,
    include_end: bool = False,
    explore_functions: bool = False,
    explore_classes: bool = False,
) -> Generator[CFGLoc, None, None]:
    yield from essor_locs_from_locs(
        [loc],
        Direction.PREDECESSORS,
        stop_on_loc,
        stop_on_block,
        include_start,
        include_end,
        explore_functions,
        explore_classes,
    )


def get_cfg_loc(in_stmt: nodes.NodeNG) -> CFGLoc:
    while not hasattr(in_stmt, "cfg_loc"):
        in_stmt = in_stmt.parent
        assert in_stmt is not None

    return in_stmt.cfg_loc


def get_stmt_locs(loc: CFGLoc) -> Tuple[Optional[CFGLoc], Optional[CFGLoc]]:
    parent = loc.node.parent

    if isinstance(loc.node, nodes.Arguments) or (
        isinstance(parent, nodes.For) and loc.node == parent.target
    ):
        return None, None

    if isinstance(parent, nodes.With):
        if loc.node == parent.items[0][0]:
            return None, parent.cfg_loc
        if any(loc.node == v for item in parent.items for v in item):
            return None, None
        return loc, None

    if (isinstance(parent, (nodes.If, nodes.While)) and loc.node == parent.test) or (
        isinstance(parent, nodes.For) and loc.node == parent.iter
    ):
        return None, parent.cfg_loc

    if isinstance(parent, nodes.Try) and loc.node == parent.body[0]:
        return loc, parent.cfg_loc

    return loc, None


def syntactic_children_locs(
    syntactic: nodes.NodeNG,
    include_stmt_locs: bool = False,
    explore_functions: bool = False,
    explore_classes: bool = False,
):
    if isinstance(syntactic, nodes.NodeNG):
        node = syntactic
    else:
        node = syntactic[0]

    assert hasattr(node, "cfg_loc")

    yield from syntactic_children_locs_from(
        node.cfg_loc, syntactic, include_stmt_locs, explore_functions, explore_classes
    )


def syntactic_children_locs_from(
    loc: CFGLoc,
    syntactic: nodes.NodeNG,
    include_stmt_locs: bool = False,
    explore_functions: bool = False,
    explore_classes: bool = False,
) -> Generator[CFGLoc, None, None]:

    if isinstance(loc, CFGLoc):
        loc = [loc]
    if isinstance(syntactic, nodes.NodeNG):
        syntactic = [syntactic]

    syntactic = [n.cfg_loc for n in syntactic]

    for is_in, block, from_pos, to_pos in get_locs_in_and_after_from(loc, syntactic):
        if is_in:
            for i in range(from_pos, to_pos):
                loc = block.locs[i]
                if not include_stmt_locs:
                    locs = [loc]
                else:
                    locs = [stmt_loc for stmt_loc in get_stmt_locs(loc) if stmt_loc is not None]

                for loc in locs:
                    yield loc

                    if isinstance(loc.node, nodes.FunctionDef) and explore_functions:
                        functions = [loc.node]
                    elif isinstance(loc.node, nodes.ClassDef) and explore_classes:
                        functions = [n for n in loc.node.body if isinstance(n, nodes.FunctionDef)]
                    else:
                        functions = []

                    for function in functions:
                        yield from successors_from_loc(
                            function.args.cfg_loc,
                            include_start=True,
                            explore_functions=explore_functions,
                            explore_classes=explore_classes,
                        )


def get_first_locs_after(locs: List[CFGLoc]):
    for is_in, block, from_pos, to_pos in get_locs_in_and_after(locs):
        if not is_in:
            yield block.locs[from_pos]


def get_locs_in_and_after(locs: List[CFGLoc]) -> Iterator[Tuple[bool, CFGBlock, int, int]]:
    yield from get_locs_in_and_after_from(locs, locs)


def _get_conseq_partition(block_locs, block):
    same_block_locs = block_locs.get(block, [])

    conseqs = []
    conseq = []

    for loc in same_block_locs:
        if len(conseq) == 0 or conseq[-1].pos + 1 == loc.pos:
            conseq.append(loc)
        else:
            conseqs.append(conseq)
            conseq = [loc]

    if len(conseq) > 0:
        conseqs.append(conseq)

    return conseqs


def get_locs_in_and_after_from(
    froms: List[CFGLoc], locs: List[CFGLoc]
) -> Iterator[Tuple[bool, CFGBlock, int, int]]:
    loc_node_set = set(s.node for s in locs)

    def stop_on(block, from_pos, to_pos):
        # it is enough to check last; locs_after inside a loc's block
        # are handled separately
        last = block.locs[to_pos - 1].node
        return len(loc_node_set & (set(last.node_ancestors()) | {last})) == 0

    block_froms = defaultdict(list)
    for loc in sorted(froms, key=lambda loc: loc.pos):
        block_froms[loc.block].append(loc)

    block_locs = defaultdict(list)
    for loc in sorted(locs, key=lambda loc: loc.pos):
        unwrapped = unwrap(loc)
        if unwrapped == loc:
            block_locs[loc.block].append(loc)
            continue

        for i in range(unwrapped.pos, len(unwrapped.block.locs)):
            subloc = unwrapped.block.locs[i]
            assert subloc.node != loc.node
            if loc.node in subloc.node.node_ancestors():
                block_locs[loc.block].append(subloc)
            else:
                break

    for succ_block, from_pos, to_pos in successor_blocks_from_locs(
        [same_block_locs[0] for same_block_locs in block_froms.values()],
        stop_on_block=stop_on,
        include_start=True,
        include_end=True,
    ):
        stop = stop_on(succ_block, from_pos, to_pos)
        conseqs = _get_conseq_partition(block_locs, succ_block)

        if len(conseqs) == 0 or (
            len(conseqs) == 1 and conseqs[0][-1].pos + 1 >= len(succ_block.locs)
        ):
            yield not stop, succ_block, from_pos, to_pos
            continue

        for i, conseq in enumerate(conseqs):
            yield True, succ_block, conseq[0].pos, conseq[-1].pos + 1

            if i < len(conseqs) - 1:
                next = conseqs[i + 1]
                yield False, succ_block, conseq[-1].pos + 1, next[0].pos
            if i == len(conseqs) - 1 and conseq[-1].pos + 1 < len(succ_block.locs):
                yield False, succ_block, conseq[-1].pos + 1, len(succ_block.locs)



from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List, Union, Dict, Optional

from astroid import nodes


ScopeNode = Union[
    nodes.Module,
    nodes.FunctionDef,
    nodes.ClassDef,
    nodes.Lambda,
    nodes.GeneratorExp,
    nodes.ListComp,
    nodes.DictComp,
    nodes.SetComp,
]
VarName = str


@dataclass(frozen=True)
class Variable:
    name: VarName
    scope: ScopeNode


class VarEventType(Enum):
    ASSIGN = auto()
    REASSIGN = auto()
    MODIFY = auto()
    READ = auto()
    DELETE = auto()


@dataclass
class VarEvent:
    var: Variable
    node: nodes.NodeNG
    type: VarEventType
    definitions: Optional[List["VarEvent"]] = field(default_factory=list)
    uses: Optional[List["VarEvent"]] = field(default_factory=list)

    def __repr__(self) -> str:
        return (
            "VarEvent("
            f"node={self.node!r}, "
            f"var={self.var.name}, "
            f"type={self.type.name}, "
            f"len(definitions)={len(self.definitions)}, "
            f"len(uses)={len(self.uses)}"
            ")"
        )

    def is_direct_modify(self):
        """Method called on the var, e.g., append, insert"""
        assert self.type == VarEventType.MODIFY
        return isinstance(self.node.parent, nodes.Attribute) and isinstance(
            self.node.parent.parent, nodes.Call
        )


@dataclass
class VarEvents:
    var_events: Dict[Variable, List[VarEvent]] = field(default_factory=lambda: defaultdict(list))

    def all(self):
        for var, events in self.var_events.items():
            for event in events:
                yield var, event

    def for_var(self, var: Variable):
        return self.var_events[var]

    def for_name(self, varname: VarName):
        for var, events in self.var_events.items():
            if var.name == varname:
                for event in events:
                    yield var, event

    def add(self, var: Variable, event: VarEvent):
        self.var_events[var].append(event)

    def __iter__(self):
        return iter(self.var_events)

    def __contains__(self, needle) -> bool:
        return needle in self.var_events

    def __getattr__(self, attr: str):
        return getattr(self.var_events, attr)

    def __getitem__(self, key):
        return self.var_events[key]


def strip_to_name(
    node: nodes.NodeNG,
) -> Optional[Union[nodes.Name, nodes.AssignName, nodes.DelName]]:
    while True:
        if isinstance(node, nodes.Subscript):
            node = node.value
        elif type(node) in (nodes.Attribute, nodes.AssignAttr, nodes.DelAttr):
            node = node.expr
        else:
            break

    return node if isinstance(node, (nodes.Name, nodes.AssignName, nodes.DelName)) else None


def unstrip(
    node: Union[nodes.Name, nodes.AssignName, nodes.DelName],
) -> Union[
    nodes.Name,
    nodes.AssignName,
    nodes.DelName,
    nodes.Subscript,
    nodes.Attribute,
    nodes.AssignAttr,
    nodes.DelAttr,
]:
    while isinstance(
        node.parent, (nodes.Subscript, nodes.Attribute, nodes.AssignAttr, nodes.DelAttr)
    ):
        node = node.parent
    return node



import json
from pathlib import Path
from typing import Optional, Dict, Any
import os
from time import sleep

from pylint.lint import PyLinter
import tomli
import tomli_w
import requests
from tqdm import tqdm

from thonny_process_slim import make_explanation_more_friedly

# This doesn't work, but we need to version the code somewhere, so we might as well do it now.
# pylint/doc/exts/pylint_messages is not in packaged pylint, but it's in the repo
# from pylint_messages import _register_all_checkers_and_extensions, _get_all_messages, _get_message_data_path

# _get_all_messages has to be modified:
# getattr(message, "shared", False),
# getattr(message, "default_enabled", False),


SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
PYLINT_EXPORT_JSON = os.path.join(SCRIPT_PATH, "pylint_export.json")
PYLINT_EXPORT_TOML = os.path.join(SCRIPT_PATH, "pylint_export.toml")
EDULINT_TOML = os.path.join(SCRIPT_PATH, "../edulint_pylint.toml")


class MyEncoder(json.JSONEncoder):
    def default(self, o):
        return o.__dict__


def read_file_or_empty_string(filepath: str) -> str:
    if not Path(filepath).is_file():
        return ""

    with open(filepath, encoding="utf8") as f:
        return f.read()


def pylint_to_json(output_filename: str):
    linter = PyLinter()
    _register_all_checkers_and_extensions(linter)
    messages, _ = _get_all_messages(linter)

    for severity in messages:
        problems = messages[severity]
        for i in range(len(problems)):
            data_path = _get_message_data_path(problems[i].definition)

            problems[i] = problems[i]._asdict()
            problems[i]["checker_module_path"] = "REDACTED"
            problems[i]["good_code"] = read_file_or_empty_string(data_path / "good.py")
            problems[i]["bad_code"] = read_file_or_empty_string(data_path / "bad.py")

    with open(output_filename, "w", encoding="utf8") as f:
        json.dump(messages, f, indent=4, cls=MyEncoder)


def convert_json_to_toml(input_filename: str, output_filename: str):
    with open(input_filename, encoding="utf8") as f:
        pylint_data = json.load(f)

    answer = {}
    for severity in pylint_data:
        for checker in pylint_data[severity]:
            answer[checker["id"]] = checker

            checker["level"] = severity
            checker["description"] = checker["definition"]["description"]
            checker["msg"] = checker["definition"]["msg"]
            del checker["definition"]

            del checker["shared"]  # This prop is unreliable
            del checker["default_enabled"]  # This prop is unreliable

    with open(output_filename, "wb") as f:
        tomli_w.dump(answer, f, multiline_strings=True)


def md_code_block_with_headline(code: str, headline: Optional[str] = None) -> str:
    if "# This is a placeholder for correct code for this message." in code:
        return ""

    answer = ""
    if headline:
        answer += f"\n**{headline}**\n"

    answer += f"""
```py
{code}
```
"""
    return answer


def produce_doc_link_if_available(checker: Dict[str, Any], skip_online_check: bool = True) -> str:
    non_viable_text = (
        '"**Additional details:**\n\nYou can help us make the doc better `by contributing'
    )
    if not checker["details"] or checker["details"].startswith(non_viable_text):
        return ""

    doc_url = f'https://pylint.pycqa.org/en/latest/user_guide/messages/{checker["level"]}/{checker["name"]}.html'
    if not skip_online_check:
        resp = requests.get(doc_url)
        sleep(0.5)
        if resp.status_code != 200:
            return ""

    return f'\n<a href="{doc_url}">Additional information</a>'


def convert_pylint_toml_to_edulint_toml(input_filename: str, output_filename: str):
    with open(input_filename, "rb") as f:
        data = tomli.load(f)

    edulint_data = {}

    for checker_id, checker in tqdm(data.items()):
        examples = ""
        if checker["bad_code"]:
            examples += md_code_block_with_headline(checker["bad_code"], "Problematic code")
        if checker["good_code"]:
            examples += md_code_block_with_headline(checker["good_code"], "How to fix it")

        examples += produce_doc_link_if_available(checker)

        edulint_data[checker_id] = {
            "why": make_explanation_more_friedly(checker["description"]),
            "examples": examples.lstrip(),
        }

    with open(output_filename, "wb") as f:
        tomli_w.dump(edulint_data, f, multiline_strings=True)


def process_from_stored_data():
    convert_json_to_toml(PYLINT_EXPORT_JSON, PYLINT_EXPORT_TOML)
    convert_pylint_toml_to_edulint_toml(PYLINT_EXPORT_TOML, EDULINT_TOML)


if __name__ == "__main__":
    # pylint_to_json(PYLINT_EXPORT_JSON)
    process_from_stored_data()



from edulint.config.arg import ProcessedArg, UnprocessedArg, ImmutableArg
from edulint.linters import Linter
from edulint.options import (
    UnionT,
    ImmutableT,
    Option,
    BASE_CONFIG,
)
from edulint.option_parses import (
    OptionParse,
    get_option_parses,
    TakesVal,
    Combine,
    OPTION_SETS_LABEL,
    LANG_TRANSLATIONS_LABEL,
    DEFAULT_ENABLER_LABEL,
)
from edulint.config.file_config import load_toml_file, get_path_relative_to
from edulint.config.option_sets import OptionSets, OptionSet, parse_option_sets
from edulint.config.language_translations import (
    LangTranslations,
    parse_lang_translations,
    parse_lang_file,
)
from edulint.config.utils import print_invalid_type_message, config_file_val_to_str, add_enabled
from typing import Dict, List, Optional, Tuple, Iterator, Any

from dataclasses import dataclass
from argparse import Namespace
from pathlib import Path
import re
import shlex
from loguru import logger
import os


class Config:
    config: List[Optional[ProcessedArg]]

    def _convert(self, args: List[UnprocessedArg]) -> List[ProcessedArg]:
        return [
            ProcessedArg(arg.option, self.option_parses[arg.option].convert(arg.val))
            for arg in args
        ]

    def _translate(
        self,
        args: List[Optional[ProcessedArg]],
        option_sets: OptionSets,
        log_unknown_groups: bool,
    ) -> List[Optional[ProcessedArg]]:
        def apply_option_set(result: List[Optional[ProcessedArg]], option_set: OptionSet) -> None:
            for linter, vals in option_set.to.items():
                translated_option = linter.to_option()
                parse = self.option_parses[translated_option]
                for translated in vals:
                    result.append(ProcessedArg(translated_option, parse.convert(translated)))

        result: List[Optional[ProcessedArg]] = []
        for arg in args:
            if arg is None:
                continue
            result.append(arg)

            if arg.option == Option.SET_GROUPS:
                for group in arg.val:
                    option_set = option_sets.get(group)
                    if option_set is None:
                        if log_unknown_groups:
                            logger.warning(
                                "unknown group {group}, known groups are {groups}",
                                group=group,
                                groups=", ".join(option_sets.keys()),
                            )
                        continue
                    apply_option_set(result, option_set)

            option_option_set = option_sets.get(arg.option.to_name())
            if option_option_set is not None:
                apply_option_set(result, option_option_set)

        return result

    def _combine(
        self, args: List[Optional[ProcessedArg]], allowed_combines: Tuple[Combine, ...] = ()
    ) -> List[Optional[ProcessedArg]]:
        indices: Dict[Option, int] = {}
        results: List[Optional[ProcessedArg]] = []
        for new_arg in args:
            if new_arg is None:
                continue
            parse = self.option_parses[new_arg.option]
            if parse.combine in allowed_combines and new_arg.option in indices:
                old_index = indices[new_arg.option]
                old_arg = results[old_index]
                assert old_arg is not None

                results.append(
                    ProcessedArg(new_arg.option, parse.combine(old_arg.val, new_arg.val))
                )
                results[old_index] = None
            else:
                results.append(new_arg)
            indices[new_arg.option] = len(results) - 1

        return results

    def __init__(
        self,
        enabler: Optional[str],
        config: Optional[List[UnprocessedArg]] = None,
        option_parses: Dict[Option, OptionParse] = get_option_parses(),
    ) -> None:
        config = config if config is not None else []
        self.option_parses = option_parses

        converted = self._convert(config)
        self.config = self._combine(converted, allowed_combines=(Combine.REPLACE,))  # type: ignore

        self.enablers = {}
        if enabler is not None:
            for arg in self.config:
                if arg is not None:
                    if isinstance(arg.val, list):
                        for val in arg.val:
                            add_enabled(enabler, self.enablers, arg.option, val)
                    elif isinstance(arg.val, str):
                        add_enabled(enabler, self.enablers, arg.option, arg.val)

    @staticmethod
    def combine(lt: "Config", rt: "Config") -> "Config":
        assert lt.option_parses == rt.option_parses
        new = Config(enabler=None, option_parses=lt.option_parses)
        new.config = new._combine(lt.config + rt.config, allowed_combines=(Combine.REPLACE,))
        new.enablers = {**lt.enablers, **rt.enablers}
        return new

    def __str__(self) -> str:
        return f"Config({', '.join(arg.option.name + '=' + str(arg.val) for arg in self.config if arg is not None)})"

    def get_last_value(self, option: Option, use_default: bool) -> Optional[UnionT]:
        assert self.option_parses[option].combine == Combine.REPLACE
        for arg in reversed(self.config):
            if arg is not None and arg.option == option:
                return arg.val
        return None if not use_default else self.option_parses[option].default

    @staticmethod
    def _to_immutable(val: UnionT) -> ImmutableT:
        return val if not isinstance(val, list) else tuple(val)

    def to_immutable(
        self, option_sets: OptionSets = {}, log_unknown_groups: bool = True
    ) -> "ImmutableConfig":
        translated = self._translate(self.config, option_sets, log_unknown_groups)
        combined = self._combine(translated, allowed_combines=(Combine.REPLACE, Combine.EXTEND))

        ordered_args = [ProcessedArg(o, self.option_parses[o].default) for o in Option]
        for arg in combined:
            if arg is None:
                continue
            ordered_args[int(arg.option)] = arg

        enablers_from_option_sets = {}
        for name, option_set in option_sets.items():
            for linter, vals in option_set.to.items():
                for val in vals:
                    add_enabled(name, enablers_from_option_sets, linter.to_option(), val)

        return ImmutableConfig(
            tuple(ImmutableArg(o, self._to_immutable(ordered_args[int(o)].val)) for o in Option),
            hashabledict({**enablers_from_option_sets, **self.enablers}),
        )


class hashabledict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


@dataclass(frozen=True)
class ImmutableConfig:
    config: Tuple[ImmutableArg, ...]
    enablers: Dict[str, str]

    def __str__(self) -> str:
        return (
            "ImmutableConfig(\n"
            + "".join(f"  {arg.option.name}={str(arg.val)}\n" for arg in self.config)
            + ")"
        )

    def __getitem__(self, option: Option) -> ImmutableT:
        return self.config[int(option)].val

    def __contains__(self, option: Option) -> bool:
        return self[option] is not None

    def __iter__(self) -> Iterator[ImmutableArg]:
        return filter(lambda x: x is not None, self.config.__iter__())


# %% components


def extract_args(path: str) -> List[str]:
    edulint_re = re.compile(r"\s*#[\s#]*edulint:\s*", re.IGNORECASE)
    ib111_re = re.compile(r".*from\s+ib111\s+import.+week_(\d+)", re.IGNORECASE)

    result: List[str] = []
    with open(path, encoding="utf-8") as f:
        for i, line in enumerate(f):
            line = line.strip()

            edmatch = edulint_re.match(line)
            if edmatch:
                raw_args = line[edmatch.end() :]
                result.extend(shlex.split(raw_args))

            ibmatch = ib111_re.match(line)
            if ibmatch:
                result.append(f"{Option.CONFIG_FILE.to_name()}=ib111.toml")

    return result


def parse_option(
    option_parses: Dict[Option, OptionParse],
    name: str,
    val: Optional[str],
) -> Optional[Option]:
    option = Option.safe_from_name(name)

    if option is None:
        logger.warning("unrecognized option {name}", name=name)
    else:
        option_parse = option_parses[option]
        if option_parse.takes_val == TakesVal.YES and val is None:
            logger.warning("option {name} takes an argument but none was supplied", name=name)
        elif option_parse.takes_val == TakesVal.NO and val is not None:
            logger.warning(
                "option {name} takes no argument but {val} was supplied", name=name, val=val
            )
        else:
            return option
    return None


def parse_args(args: List[str], option_parses: Dict[Option, OptionParse]) -> List[UnprocessedArg]:
    def get_name_val(arg: str) -> Tuple[str, Optional[str]]:
        if "=" in arg:
            name, val = arg.split("=", 1)
            return name, val
        return arg, None

    result: List[UnprocessedArg] = []
    for arg in args:
        name, val = get_name_val(arg)
        option = parse_option(option_parses, name, val)
        if option is not None:
            result.append(UnprocessedArg(option, val))
    return result


def fill_in_val(arg: UnprocessedArg, option_set: List[str]) -> List[str]:
    result = []
    for t in option_set:
        if "<val>" in t:
            assert isinstance(arg.val, str)
            result.append(t.replace("<val>", arg.val))
        else:
            result.append(t)
    return result


# %% partial parsers


def parse_cmd_config(args: List[str], option_parses: Dict[Option, OptionParse]) -> Config:
    parsed = parse_args(args, option_parses)
    return Config("cmd", parsed, option_parses)


def resolve_relative_option_path(path: str, config: Config, option: Option) -> None:
    config_path = config.get_last_value(option, use_default=False)
    if config_path is None:
        return

    assert config.option_parses[option].combine == Combine.REPLACE
    config_abspath = get_path_relative_to(config_path, path)
    config.config.append(ProcessedArg(option, config_abspath))


def resolve_ignore_infile_config(parsed: List[UnprocessedArg], config: Config) -> Config:
    if not _ignore_infile(config):
        return config

    ignore_infile_val = config.get_last_value(Option.IGNORE_INFILE_CONFIG_FOR, use_default=False)
    assert ignore_infile_val is not None
    ignore_infile_str = ",".join(ignore_infile_val)

    if "edulint" in ignore_infile_val or "all" in ignore_infile_val:
        parsed = [UnprocessedArg(Option.IGNORE_INFILE_CONFIG_FOR, ignore_infile_str)]
    else:
        parsed.append(UnprocessedArg(Option.IGNORE_INFILE_CONFIG_FOR, ignore_infile_str))

    return Config("in-file", parsed, config.option_parses)


def parse_infile_config(path: str, option_parses: Dict[Option, OptionParse]) -> Config:
    extracted = extract_args(path)
    parsed = parse_args(extracted, option_parses)
    infile_config = Config("in-file", parsed, option_parses)

    resolve_relative_option_path(path, infile_config, Option.CONFIG_FILE)
    resolve_relative_option_path(path, infile_config, Option.LANGUAGE_FILE)

    return resolve_ignore_infile_config(parsed, infile_config)


def parse_config_file(
    path: str, option_parses: Dict[Option, OptionParse]
) -> Optional[Tuple[Config, OptionSets, LangTranslations]]:
    def parse_base_config(config_dict: Dict[str, Any]) -> Optional[Config]:
        rec_config = config_dict.get(Option.CONFIG_FILE.to_name(), BASE_CONFIG)
        if not isinstance(rec_config, str):
            print_invalid_type_message(Option.CONFIG_FILE, rec_config)
            rec_config = BASE_CONFIG
        rec_config_abspath = get_path_relative_to(rec_config, path)
        return parse_config_file(rec_config_abspath, option_parses)

    config_dict = load_toml_file(path)
    if config_dict is None:
        return None

    base = (
        parse_base_config(config_dict)
        if path != BASE_CONFIG
        else _get_default_config(option_parses)
    )
    if base is None:
        return None
    base_config, base_option_sets, base_lang_tranlations = base

    result = []
    this_file_option_sets = {}
    this_file_lang_translations = {}
    for name, val in config_dict.items():
        if name == OPTION_SETS_LABEL:
            this_file_option_sets = parse_option_sets(val)
        elif name == LANG_TRANSLATIONS_LABEL:
            this_file_lang_translations = parse_lang_translations(val)
        elif name == DEFAULT_ENABLER_LABEL:
            continue
        else:
            option = parse_option(option_parses, name, val)
            if (
                option is None or option == Option.CONFIG_FILE
            ):  # config is handled as the first option
                continue

            if option == Option.LANGUAGE_FILE:
                val = get_path_relative_to(val, path)

            if not isinstance(val, dict):
                to_process = [val]
            elif option.to_name() not in {linter.to_name() for linter in Linter}:
                print_invalid_type_message(option, val)
                continue
            else:
                to_process = list(val.items())

            for val in to_process:
                str_val = config_file_val_to_str(option, val)
                if str_val is not None:
                    result.append(UnprocessedArg(option, str_val))

    enabler_name = config_dict.get(DEFAULT_ENABLER_LABEL, Path(path).stem)
    this_file_config = Config(enabler_name, result, option_parses)
    return (
        Config.combine(base_config, this_file_config),
        {**base_option_sets, **this_file_option_sets},
        {**base_lang_tranlations, **this_file_lang_translations},
    )


# %% complete parsers


def get_config_one(
    filename: str,
    cmd_args: List[str],
    option_parses: Dict[Option, OptionParse] = get_option_parses(),
) -> Optional[Tuple[ImmutableConfig, LangTranslations]]:
    configs = get_config_many([filename], cmd_args, option_parses)
    if len(configs) == 0:
        return None
    _filenames, config, lang_translations = configs[0]
    return config, lang_translations


def _partition(
    filenames: List[str], configs: List[Config], filenames_mapping: Dict[str, str]
) -> List[Tuple[List[str], Config]]:
    immutable_configs = [
        (c.to_immutable(log_unknown_groups=False), filenames_mapping[filenames[i]])
        for i, c in enumerate(configs)
    ]

    indices: Dict[ImmutableConfig, int] = {}
    partition: List[Tuple[List[str], Config]] = []
    for i, filename in enumerate(filenames):
        id_ = immutable_configs[i]
        config = configs[i]
        index = indices.get(id_)
        if index is None:
            indices[id_] = len(partition)
            partition.append(([filename], config))
        else:
            partition[index][0].append(filename)
    return partition


def _get_default_config(option_parses: Dict[Option, OptionParse]) -> Tuple[Config, OptionSets]:
    return Config(enabler=None, option_parses=option_parses), {}, {}


def _ignore_infile(config: Config) -> bool:
    ignored_infile = config.get_last_value(Option.IGNORE_INFILE_CONFIG_FOR, use_default=True)
    return Linter.EDULINT.to_name() in ignored_infile or "all" in ignored_infile


# def get_config_paths(
#     filenames: List[str], infile_configs: List[Config]
# ) -> Tuple[Set[str], Dict[str, str]]:
#     assert len(filenames) == len(infile_configs)
#     config_paths = set()
#     filename_mapping = {}
#     for filename, config in zip(filenames, infile_configs):
#         config_file = config.get_last_value(Option.CONFIG_FILE, use_default=True)
#         if (
#             config_file.startswith("http")
#             or not config_file.endswith(".toml")
#             or os.path.isabs(config_file)
#         ):
#             config_path = config_file
#         else:
#             config_path = str(Path(filename).parent / config_file)
#         config_paths.add(config_path)
#         filename_mapping[filename] = config_path
#     return config_paths, filename_mapping


def _parse_infile_configs(
    files_or_dirs: List[str], cmd_config: Config, option_parses: Dict[Option, OptionParse]
) -> Dict[ImmutableConfig, Tuple[Config, List[str]]]:

    def add_to_result(result, config, path, iconfig=None):
        iconfig = config.to_immutable() if iconfig is None else iconfig
        if iconfig not in result:
            result[iconfig] = (config, [path])
        else:
            result[iconfig][1].append(path)

    def aggregate_subresults(paths):
        result = {}
        for path in paths:
            subresult = _parse_infile_configs_rec(path)
            for iconfig, (config, subpaths) in subresult.items():
                if iconfig not in result:
                    result[iconfig] = (config, subpaths)
                else:
                    result[iconfig][1].extend(subpaths)
        return result

    def _parse_infile_configs_rec(path: str):
        if not os.path.isdir(path):
            if not os.path.splitext(path)[1].lower() == ".py":
                return {}

            infile_config = parse_infile_config(path, option_parses)
            return {infile_config.to_immutable(log_unknown_groups=False): (infile_config, [path])}

        result = aggregate_subresults([os.path.join(path, name) for name in os.listdir(path)])

        if len(result) != 1:
            return result

        iconfig, (config, _paths) = list(result.items())[0]
        return {iconfig: (config, [path])}

    if _ignore_infile(cmd_config):
        default_config = _get_default_config(option_parses)[0]
        return {default_config.to_immutable(): (default_config, files_or_dirs)}

    return aggregate_subresults(files_or_dirs)


def get_config_many(
    files_or_dirs: List[str],
    cmd_args_raw: List[str],
    option_parses: Dict[Option, OptionParse] = get_option_parses(),
) -> List[Tuple[List[str], ImmutableConfig, LangTranslations]]:
    cmd_config = parse_cmd_config(cmd_args_raw, option_parses)
    infile_configs = _parse_infile_configs(files_or_dirs, cmd_config, option_parses)

    cmd_config_path = cmd_config.get_last_value(Option.CONFIG_FILE, use_default=False)

    if cmd_config_path is not None:
        config_paths = {cmd_config_path}
    else:
        config_paths = {iconfig[Option.CONFIG_FILE] for iconfig in infile_configs.keys()}

    config_file_results = {
        config_path: parse_config_file(config_path, option_parses) for config_path in config_paths
    }
    lang_file_results = {}

    result: List[Tuple[List[str], ImmutableConfig, LangTranslations]] = []
    for infile_config, paths in infile_configs.values():
        combined = Config.combine(infile_config, cmd_config)
        config_file_result = (
            config_file_results[cmd_config_path]
            if cmd_config_path is not None
            else config_file_results[
                infile_config.get_last_value(Option.CONFIG_FILE, use_default=True)
            ]
        )
        if config_file_result is None:
            continue
        file_config, option_sets, lang_translations = config_file_result

        if _ignore_infile(file_config):
            if cmd_config_path is not None:
                config = Config.combine(file_config, cmd_config)
            else:
                config = cmd_config
        else:
            config = Config.combine(file_config, combined)

        iconfig = config.to_immutable(option_sets)
        language_file = iconfig[Option.LANGUAGE_FILE]
        if language_file is not None:
            lang_file_translations = lang_file_results.get(
                language_file, parse_lang_file(iconfig[Option.LANGUAGE_FILE])
            )
            if lang_file_translations is not None:
                lang_translations = {**lang_translations, **lang_file_translations}
                lang_file_results[language_file] = lang_translations

        result.append((paths, iconfig, lang_translations))
    return result


def get_cmd_args(args: Namespace) -> List[str]:
    return [s for arg in args.options for s in shlex.split(arg)]



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from test_utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def fun(x, y):",
        "    if x > y:",
        "        print('foo')",
        "    elif x <= y:",
        "        print('bar')"
    ], [
        lazy_problem().set_line(4).set_code("R6611")
    ]),
    ([
        "def fun(x, y):",
        "    if x > y:",
        "        print('foo')",
        "    if x <= y:",
        "        print('bar')"
    ], [
        # lazy_problem().set_line(3).set_code("R6611")  # TODO report another defect
    ]),
    ([
        "def fun(x, y):",
        "    if x > y:",
        "        print('foo')",
        "    if x <= y:",
        "        print('bar')",
        "    else:",
        "        print('foobar')"
    ], [
    ]),
])
def test_local_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=local-defects")],
        expected_output
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def foo(a):",
        "    a = a + 1",
    ], [lazy_problem()]),
    # ([
    #     "def foo(a):",
    #     "    a = a * 2",
    # ], []),
    ([
        "def foo(a):",
        "    a = a + [0]",
    ], []),
    ([
        "def foo(a, b):",
        "    a = a + b",
    ], []),
    ([
        "def foo(a):",
        "    a = [0] + a",
    ], []),
    ([
        "def foo(a):",
        "    a = a + 'a'",
    ], [lazy_problem()]),
    ([
        "def foo(a):",
        "    a = 'a' + a",
    ], []),
    ([
        "def foo(a):",
        "    a = a & 2",
    ], [lazy_problem()]),
    ([
        "def foo(a):",
        "    a = a & {'a'}",
    ], []),
    ([
        "def foo(a, b):",
        "    a = a & b",
    ], []),
    ([
        "def foo(a):",
        "    b = ''",
        "    a = a + b",
    ], [lazy_problem()]),
    ([
        "def foo(a):",
        "    b = ''",
        "    a = b + a",
    ], []),
    ([
        "def foo(a):",
        "    b = ''",
        "    b = a + b",
    ], []),
])
def test_augassign_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=use-augmented-assign")],
        [problem.set_code("R6609") for problem in expected_output]
    )

@pytest.mark.parametrize("lines,expected_output", [
    (["1 is True"], [lazy_problem()]),
    (["1 is not True"], [lazy_problem()]),
    (["1 is None"], []),
    (["1 is not None"], []),
])
def test_no_is_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=no-is-bool")],
        [problem.set_code("R6613") for problem in expected_output]
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("044050-vigenere.py", [
        lazy_problem().set_code("R6614").set_line(38)
        .set_text("Use \"ord('A')\" instead of using the magical constant 65."),
        lazy_problem().set_code("R6615").set_line(39)
        .set_text("Remove the call to 'ord' and compare to the string directly \"> 'Z'\" instead of using "
                    "the magical constant 91. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6614").set_line(40)
        .set_text("Use \"(ord('Z') + 1)\" instead of using the magical constant 91."),
        lazy_problem().set_code("R6614").set_line(68)
        .set_text("Use \"ord('A')\" instead of using the magical constant 65."),
        lazy_problem().set_code("R6615").set_line(69)
        .set_text("Remove the call to 'ord' and compare to the string directly \"< 'A'\" instead of using "
                    "the magical constant 65. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6614").set_line(70)
        .set_text("Use \"ord('A')\" instead of using the magical constant 65.")
    ]),
    ("044262-vigenere.py", [  # TODO detect also 42 and 72
        lazy_problem().set_code("R6615").set_line(36)
        .set_text("Remove the call to 'ord' and compare to the string directly \"< 'A'\" instead of using "
                    "the magical constant 65. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6615").set_line(36)
        .set_text("Remove the call to 'ord' and compare to the string directly \"> '['\" instead of using "
                    "the magical constant 91. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6615").set_line(66),
        lazy_problem().set_code("R6615").set_line(66)
        .set_text("Remove the call to 'ord' and compare to the string directly \"> '['\" instead of using "
                    "the magical constant 91. Careful, this may require changing the comparison operator."),
    ]),
    ("048426-ipv4.py", [
        lazy_problem().set_code("R6614").set_line(4)
        .set_text("Use \"ord('0')\" instead of using the magical constant 48.")
    ]),
    ("hw34017.py", [
        lazy_problem().set_code("R6615").set_line(177),
        lazy_problem().set_code("R6615").set_line(177)
        .set_text("Remove the call to 'ord' and compare to the string directly \"< 'D'\" instead of using "
                    "the magical constant 68. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6615").set_line(177),
        lazy_problem().set_code("R6615").set_line(177)
        .set_text("Remove the call to 'ord' and compare to the string directly \"< '3'\" instead of using "
                    "the magical constant 51. Careful, this may require changing the comparison operator."),
        lazy_problem().set_code("R6614").set_line(178),
    ]),
])
def test_ord_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [
            Arg(Option.PYLINT, "--enable=use-ord-letter,use-literal-letter")
        ],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("010666-prime.py", [
        lazy_problem().set_code("R6604").set_line(12).set_text("Do not use while loop with else."),
        lazy_problem().set_code("R6609").set_line(15).set_text("Use augmented assignment: 'i += 1'"),
    ]),
    ("012889-pythagorean.py", [
        lazy_problem().set_code("R6602").set_line(23).set_text("Use integral division //."),
        lazy_problem().set_code("R6602").set_line(24),
        lazy_problem().set_code("R6602").set_line(25),
    ]),
    ("014290-coins.py", [
        lazy_problem().set_code("R6608").set_line(27).set_text("Redundant arithmetic: 1 * (value // 1)")
    ]),
    ("014494-next.py", [
        lazy_problem().set_code("R6602").set_line(5).set_text("Use integral division //."),
        lazy_problem().set_code("R6608").set_line(16).set_text("Redundant arithmetic: number**1"),
    ]),
    ("014661-p5_fibsum.py", [
        lazy_problem().set_code("R6609").set_line(15),
        lazy_problem().set_code("R6609").set_line(18),
        lazy_problem().set_code("R6609").set_line(19),
        lazy_problem().set_code("R6609").set_line(20),
        lazy_problem().set_code("R6611").set_line(23).set_text("Use else instead of elif."),
        lazy_problem().set_code("R6609").set_line(24),

    ]),
    ("017667-prime.py", [
        lazy_problem().set_code("R6616").set_line(5).set_text("Use early return."),
        lazy_problem().set_code("R6604").set_line(6).set_text("Do not use for loop with else."),
    ]),
    ("022859-digit_sum.py", [
        lazy_problem().set_code("R6601").set_line(20)
        .set_text("Use lst.append(number % 7) instead of lst += [number % 7].")
    ]),
    ("024042-cards.py", [
        lazy_problem().set_code("R6609").set_line(9).set_text("Use augmented assignment: 'number //= 10'"),
        lazy_problem().set_code("R6613").set_line(37)
        .set_text("Use 'is_valid_card(number)' directly rather than as 'is_valid_card(number) is True'."),
        lazy_problem().set_code("R6613").set_line(55),
    ]),
    ("024180-delete.py", [
        lazy_problem().set_code("R6606").set_line(7)
        .set_text("The for loop makes only one iteration.")
    ]),
    ("024233-cards.py", [
        lazy_problem().set_code("R6608").set_line(22).set_text("Redundant arithmetic: 0 + number"),
        lazy_problem().set_code("R6616").set_line(23),
        lazy_problem().set_code("R6608").set_line(42).set_text("Redundant arithmetic: 0 + number"),
        lazy_problem().set_code("R6616").set_line(43),
        lazy_problem().set_code("R6616").set_line(47)
    ]),
    ("024371-cards.py", [
        lazy_problem().set_code("R6606").set_line(12).set_text("The for loop makes no iterations.")
    ]),
    ("024371-workdays.py", [
        lazy_problem().set_code("R6606").set_line(29)
        .set_text("The for loop makes only one iteration."),
        lazy_problem().set_code("R6608").set_line(30).set_text("Redundant arithmetic: i + 0"),
    ]),
    ("024491-cards.py", [
        lazy_problem().set_code("R6606").set_line(86)
        .set_text("The for loop makes only one iteration.")
    ]),
    ("024657-bisection.py", [
        lazy_problem().set_code("R6608").set_line(6).set_text("Redundant arithmetic: 0 - eps"),
        lazy_problem().set_code("R6608").set_line(6).set_text("Redundant arithmetic: 0 + eps"),
        lazy_problem().set_code("R6608").set_line(8).set_text("Redundant arithmetic: 0 - eps"),
        lazy_problem().set_code("R6608").set_line(10).set_text("Redundant arithmetic: 0 + eps"),
        lazy_problem().set_code("R6608").set_line(12).set_text("Redundant arithmetic: 0 + eps"),
        lazy_problem().set_code("R6608").set_line(14).set_text("Redundant arithmetic: 0 + eps"),
    ]),
    ("032630-concat.py", [
        lazy_problem().set_code("R6608").set_line(5).set_text("Redundant arithmetic: [0] * 0")
    ]),
    ("041630-ipv4.py", [
        lazy_problem().set_code("R6603").set_line(15).set_text("Use isdecimal to test if string contains a number.")
    ]),
    ("043232-person_id.py", [  # no use else instead of elif on line
        lazy_problem().set_code("R6613").set_line(44),
        lazy_problem().set_code("R6613").set_line(44),
    ]),
    ("044834-ipv4.py", [
        lazy_problem().set_code("R6603").set_line(15).set_text("Use isdecimal to test if string contains a number.")
    ]),
    ("064633-p4_histogram.py", [
        lazy_problem().set_code("R6611").set_line(35).set_text("Use else instead of elif.")
    ]),
    ("085286-p1_count.py", [
        lazy_problem().set_code("R6611").set_line(27).set_text("Use else instead of elif."),
        lazy_problem().set_code("R6612").set_line(30).set_text("Unreachable else."),
        lazy_problem().set_code("R6611").set_line(41).set_text("Use else instead of elif."),
        lazy_problem().set_code("R6612").set_line(44).set_text("Unreachable else."),
    ]),
    ("3ee0761302-p4_geometry.py", []),
    ("4594fad43b-p4_triangle.py", []),
    ("66479c3d7d-p4_geometry.py", []),
    ("7922cc0638-p6_next.py", [lazy_problem().set_code("R6616").set_line(20)]),
    ("104174-ipv4_restore.py", [
        lazy_problem().set_code("R6610").set_line(38).set_text("Do not multiply list with mutable content.")
    ]),
    ("104584-p4_digits.py", []),
    ("calc_with_vars.py", []),
    ("hw14358.py", [
        lazy_problem().set_code("R6609").set_line(31).set_text("Use augmented assignment: 'b += 1'"),
        lazy_problem().set_code("R6609").set_line(34).set_text("Use augmented assignment: 'a += 1'"),
        lazy_problem().set_code("R6608").set_line(44).set_text("Redundant arithmetic: num // num"),
        lazy_problem().set_code("R6609").set_line(50).set_text("Use augmented assignment: 'count += 1'"),
        lazy_problem().set_code("R6609").set_line(51).set_text("Use augmented assignment: 'number //= n'"),
        lazy_problem().set_code("R6609").set_line(53).set_text("Use augmented assignment: 'primes += 1'"),
        lazy_problem().set_code("R6609").set_line(54),
        lazy_problem().set_code("R6609").set_line(56),
        lazy_problem().set_code("R6609").set_line(57),
    ]),
    ("hw34328.py", [
        lazy_problem().set_code("R6608").set_line(79).set_text("Redundant arithmetic: '' + str(count)")
    ]),
    ("hw34451.py", [
        lazy_problem().set_code("R6616").set_line(13),
        lazy_problem().set_code("R6607").set_line(44)
        .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
        lazy_problem().set_code("R6607").set_line(47)
        .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
        lazy_problem().set_code("R6607").set_line(63)
        .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
        lazy_problem().set_code("R6607").set_line(66)
        .set_text("Use exponentiation instead of repeated muliplication in i * i * i."),
        lazy_problem().set_code("R6613").set_line(98)
        .set_text("Use 'not end' directly rather than as 'end is False'.")
    ]),
    ("hw35219.py", [
        lazy_problem().set_code("R6601").set_line(34)
        .set_text("Use coords_cw.append((-y, x)) instead of coords_cw += [(-y, x)]."),
        lazy_problem().set_code("R6601").set_line(41)
        .set_text("Use coords_ccw.append((y, -x)) instead of coords_ccw += [(y, -x)]."),
        lazy_problem().set_code("R6610").set_line(46).set_text("Do not multiply list with mutable content.")
    ]),
    ("hw45208.py", [
        lazy_problem().set_code("R6605").set_line(32),
        lazy_problem().set_code("R6611").set_line(45),
    ]),
    ("m2630.py", [
        lazy_problem().set_code("R6608").set_line(59).set_text("Redundant arithmetic: [0] * 0"),
        lazy_problem().set_code("R6608").set_line(79).set_text("Redundant arithmetic: num //= num"),
        lazy_problem().set_code("R6608").set_line(83).set_text("Redundant arithmetic: num //= num"),
        lazy_problem().set_code("R6608").set_line(86).set_text("Redundant arithmetic: num //= num"),
        lazy_problem().set_code("R6608").set_line(97).set_text("Redundant arithmetic: position * 1"),
        lazy_problem().set_code("R6613").set_line(144),
        lazy_problem().set_code("R6613").set_line(144),
    ]),
    ("m2650.py", [
        lazy_problem().set_code("R6609").set_line(11),
        lazy_problem().set_code("R6609").set_line(41),
        lazy_problem().set_code("R6609").set_line(44),
        lazy_problem().set_code("R6605").set_line(84).set_text("Use elif."),
    ]),
    ("m5435.py", [
        lazy_problem().set_code("R6609").set_line(7),
        lazy_problem().set_code("R6605").set_line(9).set_text("Use elif."),
        lazy_problem().set_code("R6608").set_line(53).set_text("Redundant arithmetic: valid += 0"),
        lazy_problem().set_code("R6608").set_line(64).set_text("Redundant arithmetic: valid += 0"),
    ]),
    ("ut_157_0762_16_47.py", []),
])
def test_local_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=local-defects")],
        expected_output
    )



from typing import Any, Dict, ClassVar, List
from dataclasses import dataclass
import inspect
import pkgutil
import importlib
import re

from loguru import logger

import pylint
from pylint.checkers import BaseChecker

from edulint.option_parses import LANG_TRANSLATED_EXTRACTS_LABEL
from edulint.config.file_config import load_toml_file
from edulint.config.raw_flake8_patterns import FLAKE8
from edulint.linting import checkers as edulint_checkers


def to_pattern(message: str):
    return re.sub("%[rsd]", "(.*)", re.escape(message), flags=re.IGNORECASE)


def checker_classes(module):
    for module_info in pkgutil.walk_packages(module.__path__):
        submodule = importlib.import_module(f"{module.__name__}.{module_info.name}")
        for name, obj in inspect.getmembers(submodule):
            if inspect.isclass(obj) and issubclass(obj, BaseChecker) and obj != BaseChecker:
                yield obj


def get_edulint_patterns():
    messages = {}
    for checker in checker_classes(edulint_checkers):
        messages.update({code: info[0] for code, info in checker.msgs.items()})

    return messages


def get_pylint_patterns():
    messages = {}
    for checker in checker_classes(pylint.checkers):
        messages.update({code: info[0] for code, info in checker.msgs.items()})

    return messages


def get_patterns():
    return {
        code: to_pattern(message)
        for code, message in {**get_pylint_patterns(), **get_edulint_patterns(), **FLAKE8}.items()
    }


@dataclass
class Translation:
    translation: str
    extracts: Dict[str, Dict[str, str]]
    patterns: ClassVar[Dict[str, str]] = get_patterns()

    def translate_extracts(self, extracts: List[str]) -> List[str]:
        result = []
        for i, word in enumerate(extracts):
            mapping = self.extracts.get(str(i + 1))
            if mapping is not None:
                result.append(mapping.get(word, word))
            else:
                result.append(word)
        return result

    def translate(self, code: str, message: str):
        pattern = self.patterns.get(code)
        if pattern is None or "(.*)" not in pattern:
            return self.translation

        match = re.match(pattern, message, flags=re.IGNORECASE)
        if not match:
            return self.translation
        return self.translation.format(*self.translate_extracts(match.groups()))


LangTranslations = Dict[str, Translation]


def parse_lang_translations(raw_lang_translations: Any) -> LangTranslations:
    if not isinstance(raw_lang_translations, dict):
        logger.warning(
            "langauge translations are not a dictionary but a value {val} of type {type}",
            val=raw_lang_translations,
            type=type(raw_lang_translations),
        )
        return None

    translated_extracts = raw_lang_translations.get(LANG_TRANSLATED_EXTRACTS_LABEL, {})
    if not isinstance(translated_extracts, dict):
        logger.warning(
            "translation for specific extracts is not a dictionary but a value of type {type}",
            type=type(translated_extracts),
        )
        translated_extracts = {}
    for id_, val in translated_extracts.items():
        if not isinstance(val, dict):
            logger.warning(
                "translated extracts for identifier {id_} is not a dictionary but a value of type {type}",
                id_=id_,
                type=type(val),
            )
            continue
        for order, mapping in val.items():
            if not order.isdecimal():
                logger.warning(
                    "order value {order} of translated extracts for identifier {id_} does not contain integer",
                    order=order,
                    id_=id_,
                )
            if not isinstance(mapping, dict):
                logger.warning(
                    "translated extracts mapping for order {order} of identifier {id_} is not a dicitonary but a value of type {type}",
                    order=order,
                    id_=id_,
                    type=type(mapping),
                )
                continue
            for extract, translated in mapping.items():
                if not isinstance(translated, str):
                    logger.warning(
                        "translation for extract {extract} for order {order} of identifier {id_} is not a string but a value of type {type}",
                        extract=extract,
                        order=order,
                        id_=id_,
                        type=type(translated),
                    )

    lang_translations = {}
    for id_, translation in raw_lang_translations.items():
        if id_ == LANG_TRANSLATED_EXTRACTS_LABEL:
            continue
        if not isinstance(translation, str):
            logger.warning(
                "translation for identifier {id_} is not a string but a value of type {type}",
                id_=id_,
                type=type(translation),
            )
            continue

        lang_translations[id_] = Translation(translation, translated_extracts.get(id_, {}))

    return lang_translations


def parse_lang_file(filename: str):
    lang_file_raw = load_toml_file(filename)
    if lang_file_raw is None:
        return {}
    return parse_lang_translations(lang_file_raw)



# adapted from https://github.com/pyta-uoft/pyta/blob/4c858623549e24a49fea7aef9c8ec7c20c836bd6/python_ta/cfg/graph.py

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Generator, List, Optional, Set

from astroid import nodes

from edulint.linting.analyses.var_events import VarEvents


class ControlFlowGraph:
    """A graph representing the control flow of a Python program."""

    start: CFGBlock
    end: CFGBlock
    # The unique id of this cfg. Defaults to 0 if not initialized in a CFGVisitor instance.
    cfg_id: int
    # block_count is used as an "autoincrement" to ensure the block ids are unique.
    block_count: int
    # blocks (with at least one statement) that will never be executed in runtime.
    unreachable_blocks: Set[CFGBlock]

    def __init__(self, cfg_id: int = 0) -> None:
        self.block_count = 0
        self.cfg_id = cfg_id
        self.unreachable_blocks = set()
        self.start = self.create_block()
        self.end = self.create_block()

    def create_block(
        self, pred: Optional[CFGBlock] = None, edge_label: Optional[Any] = None
    ) -> CFGBlock:
        """Create a new CFGBlock for this graph.

        If pred is specified, set that block as a predecessor of the new block.

        If edge_label is specified, set the corresponding edge in the CFG with that label.
        """
        new_block = CFGBlock(self.block_count)
        self.unreachable_blocks.add(new_block)

        self.block_count += 1
        if pred:
            self.link_or_merge(pred, new_block, edge_label)
        return new_block

    def link(self, source: CFGBlock, target: CFGBlock) -> None:
        """Link source to target."""
        if not source.is_jump():
            CFGEdge(source, target)

    def link_or_merge(
        self, source: CFGBlock, target: CFGBlock, edge_label: Optional[Any] = None
    ) -> None:
        """Link source to target, or merge source into target if source is empty.

        An "empty" node for this purpose is when source has no statements.

        source with a jump statement cannot be further linked or merged to
        another target.

        If edge_label is specified, set the corresponding edge in the CFG with that label.
        """
        if source.is_jump():
            return
        if source.locs == []:
            if source is self.start:
                self.start = target
            else:
                for edge in source.predecessors:
                    edge.target = target
                    target.predecessors.append(edge)
            # source is a utility block that helps build the cfg that does not
            # represent any part of the program so it is redundant.
            self.unreachable_blocks.remove(source)
        else:
            CFGEdge(source, target, edge_label)

    def multiple_link_or_merge(self, source: CFGBlock, targets: List[CFGBlock]) -> None:
        """Link source to multiple target, or merge source into targets if source is empty.

        An "empty" node for this purpose is when source has no statements.

        source with a jump statement cannot be further linked or merged to
        another target.

        Precondition:
            - source != cfg.start
        """
        if source.locs == []:
            for edge in source.predecessors:
                for t in targets:
                    CFGEdge(edge.source, t)
                edge.source.successors.remove(edge)
            source.predecessors = []
            self.unreachable_blocks.remove(source)
        else:
            for target in targets:
                self.link(source, target)

    def get_blocks(self) -> Generator[CFGBlock, None, None]:
        """Generate a sequence of all blocks in this graph."""
        stack = [(0, self.start)]
        visited = {self.start}

        while stack:
            i, block = stack.pop()
            if i == 0:
                yield block
            if i >= len(block.successors):
                continue

            stack.append((i + 1, block))

            next_block = block.successors[i].target
            if next_block in visited:
                continue

            stack.append((0, next_block))
            visited.add(next_block)

    def get_blocks_postorder(self) -> Generator[CFGBlock, None, None]:
        """Return the sequence of all blocks in this graph in the order of
        a post-order traversal."""
        yield from self._get_blocks_postorder(self.start, set())

    def _get_blocks_postorder(self, block: CFGBlock, visited) -> Generator[CFGBlock, None, None]:
        if block.id in visited:
            return

        visited.add(block.id)
        for succ in block.successors:
            yield from self._get_blocks_postorder(succ.target, visited)

        yield block

    def get_edges(self) -> Generator[CFGEdge, None, None]:
        """Generate a sequence of all edges in this graph."""
        yield from self._get_edges(self.start, set())

    def _get_edges(self, block: CFGBlock, visited: Set[int]) -> Generator[CFGEdge, None, None]:
        if block.id in visited:
            return

        visited.add(block.id)

        for edge in block.successors:
            yield edge
            yield from self._get_edges(edge.target, visited)

    def update_block_reachability(self) -> None:
        for block in self.get_blocks():
            block.reachable = True
            if block in self.unreachable_blocks:
                self.unreachable_blocks.remove(block)


@dataclass
class CFGLoc:
    block: CFGBlock
    pos: int
    node: nodes.NodeNG
    var_events: VarEvents = field(default_factory=VarEvents)

    def __eq__(self, other):
        return self.block == other.block and self.pos == other.pos and self.node == other.node

    def __hash__(self):
        return hash((self.block, self.pos, self.node))

    def __repr__(self):
        return f"CFGLoc(node={self.node!r})"


class CFGBlock:
    """A node in a control flow graph.

    Represents a maximal block of code whose statements are guaranteed to execute in sequence.
    """

    # A unique identifier
    id: int
    # The statements in this block.
    locs: List[CFGLoc]
    # This block's in-edges (from blocks that can execute immediately before this one).
    predecessors: List[CFGEdge]
    # This block's out-edges (to blocks that can execute immediately after this one).
    successors: List[CFGEdge]
    # Whether there exists a path from the start block to this block.
    reachable: bool

    def __init__(self, id_: int) -> None:
        """Initialize a new CFGBlock."""
        self.id = id_
        self.locs = []
        self.predecessors = []
        self.successors = []
        self.reachable = False

    def __repr__(self):
        return f"CFGBlock(len={len(self.locs)}, fst={self.locs[0].node!r})"

    def add_statement(self, statement: nodes.NodeNG) -> None:
        if not self.is_jump():
            loc = CFGLoc(self, len(self.locs), statement)
            self.locs.append(loc)
            statement.cfg_loc = loc

    def _CF_statement_to_pos(self, statement: nodes.NodeNG) -> int:
        if isinstance(statement, nodes.Module):
            pos = 0
            # assert self.locs[pos] == self.locs[self._CF_statement_to_pos(statement.body[0])]
        elif isinstance(statement, nodes.If):
            pos = len(self.locs) - 1
            assert self.locs[pos].node == statement.test
        elif isinstance(statement, nodes.While):
            pos = 0
            assert self.locs[pos].node == statement.test
        elif isinstance(statement, nodes.For):
            pos = len(self.locs) - 1
            assert self.locs[pos].node == statement.iter
        elif isinstance(statement, (nodes.Try, nodes.TryStar)):
            pos = 0
            statement_loc = statement.body[0].cfg_loc
            assert self.locs[pos].node == statement_loc.block.locs[statement_loc.pos].node
        elif isinstance(statement, nodes.ExceptHandler):
            pos = 0
            first_body_loc = statement.body[0].cfg_loc
            first_body_node = first_body_loc.block.locs[first_body_loc.pos].node
            assert self.locs[pos].node in (statement.name, first_body_node)
        elif isinstance(statement, nodes.With):
            pos = len(self.locs)
        else:
            assert False, f"unreachable, but {type(statement)}"
        return pos

    def add_CF_statement(self, statement: nodes.NodeNG) -> None:
        pos = self._CF_statement_to_pos(statement)
        loc = CFGLoc(self, pos, statement)
        statement.cfg_loc = loc

    @property
    def jump(self) -> Optional[nodes.NodeNG]:
        if len(self.locs) > 0:
            return self.locs[-1]

    def is_jump(self) -> bool:
        """Returns True if the block has a statement that branches
        the control flow (ex: `break`)"""
        return isinstance(self.jump, (nodes.Break, nodes.Continue, nodes.Return, nodes.Raise))


class CFGEdge:
    """An edge in a control flow graph.

    Edges are directed, and in the future may be augmented with auxiliary metadata about the control flow.
    """

    source: CFGBlock
    target: CFGBlock
    label: Optional[Any]

    def __init__(
        self, source: CFGBlock, target: CFGBlock, edge_label: Optional[Any] = None
    ) -> None:
        self.source = source
        self.target = target
        self.label = edge_label
        self.source.successors.append(self)
        self.target.predecessors.append(self)



from typing import Dict, Set


OVERRIDERS = {
    # overriden: { overriders }
    "C0104": {"R6307", "R6609", "F841", "W0613"},
    # C0104=diallowed-name; R6307=use-for-each; R6609=augmenting-assignment
    #                       F841=unused variable; W0613=unused-argument
    "C0103": {"R6307", "R6609", "F841", "W0613"},
    # C0103=invalid-name
    "E225": {"R6609"},
    # E225=missing whitespace around operator
    "R1705": {"R1703", "R6201"},
    # R1705=no-else-return, R1703=simplifiable-if-statement
    "R6610": {"R6608"},
    # R6610=do-not-multiply-mutable, R6608=redundant-arithmetic
    "W0107": {"W0101"},
    # W0107=unnecessary-pass, W0101=unreachable
    "C0201": {"C0206"},
    # C0201=consider-iterating-dictionary, C0206=consider-using-dict-items
}


def get_overriders() -> Dict[str, Set[str]]:
    return OVERRIDERS



import pytest
import os
from edulint.linters import Linter
from edulint.config.arg import UnprocessedArg, ProcessedArg
from edulint.options import Option, DEFAULT_CONFIG
from edulint.option_parses import OptionParse, get_option_parses, TakesVal, Type, Combine
from edulint.config.arg import Arg
from edulint.config.config import (
    Config,
    extract_args,
    parse_args,
    parse_config_file,
    get_config_many,
    get_config_one,
)
from edulint.config.option_sets import OptionSet, OptionSets
from edulint.linting.tweakers import get_tweakers
from test_utils import get_tests_path, remote_empty_config_url
from typing import List, Set, Dict, Tuple, Optional
from pathlib import Path


@pytest.fixture
def all_advertised_options() -> List[Option]:
    return list(get_option_parses().keys())


@pytest.fixture
def advertised_options(all_advertised_options: List[Option]) -> Set[Option]:
    return set(all_advertised_options)


@pytest.fixture
def always_managed_options() -> Set[Option]:
    return set(
        (
            Option.CONFIG_FILE,
            Option.PYLINT,
            Option.FLAKE8,
            Option.NO_FLAKE8,
            Option.IGNORE_INFILE_CONFIG_FOR,
            Option.EXPORT_GROUPS,
            Option.SET_GROUPS,
            Option.LANGUAGE_FILE,
        )
    )


@pytest.fixture
def managed_tweaker_options() -> List[Set[Option]]:
    return list(tweaker.used_options for tweaker in get_tweakers().values())


@pytest.fixture
def managed_options(
    always_managed_options: Set[Option], managed_tweaker_options: List[Set[Option]]
) -> Set[Option]:
    return always_managed_options | set(
        opt for options in managed_tweaker_options for opt in options
    )


def issubset(subset, superset):
    diff = subset - superset
    assert not diff


def test_no_duplicate_advertised_options(all_advertised_options, advertised_options):
    issubset(set(all_advertised_options), advertised_options)


def test_all_advertised_options_managed(
    advertised_options: List[Option], managed_options: Set[Option]
):
    issubset(advertised_options, managed_options)


def test_all_managed_options_are_advertised(
    advertised_options: List[Option], managed_options: Set[Option]
) -> None:
    issubset(managed_options, advertised_options)


def mock_contents(mocker, contents) -> None:
    mocked_file = mocker.mock_open(read_data=contents)
    mocker.patch("builtins.open", mocked_file)


@pytest.mark.parametrize(
    "contents,args",
    [
        ("", []),
        ('print("Hello world")', []),
        ("# edulint: xxx", ["xxx"]),
        ("# edulint xxx", []),
        ("# edulint: pylint=xxx", ["pylint=xxx"]),
        ("# edulint: flake8=xxx", ["flake8=xxx"]),
        ("# edulint: xxx yyy", ["xxx", "yyy"]),
        ("# edulint: xxx\n# edulint: xxx", ["xxx", "xxx"]),
        ("# edulint: xxx\n# edulint: yyy", ["xxx", "yyy"]),
        ("# edulint: pylint=xxx\n# edulint: pylint=yyy", ["pylint=xxx", "pylint=yyy"]),
        ("# edulint: xxx\n# edulint: pylint=yyy", ["xxx", "pylint=yyy"]),
        ("# edulint: flake8=xxx\n# edulint: pylint=yyy", ["flake8=xxx", "pylint=yyy"]),
        ("\n\n# edulint: xxx", ["xxx"]),
        (
            "some code\n # edulint: xxx\ncode mentioning edulint a bit\n# edulint: yyy",
            ["xxx", "yyy"],
        ),
        ("           #        edulint:      xxx        ", ["xxx"]),
        ("           #        edulint:      xxx  yyy      ", ["xxx", "yyy"]),
        ('# edulint: pylint="xxx       yyy"', ["pylint=xxx       yyy"]),
        (
            '# edulint: pylint=xxx yyy zzz flake8=aaa bbb pylint= flake8="xxx yyy" ',
            ["pylint=xxx", "yyy", "zzz", "flake8=aaa", "bbb", "pylint=", "flake8=xxx yyy"],
        ),
        (
            "# edulint: pylint=--enable=missing-module-docstring",
            ["pylint=--enable=missing-module-docstring"],
        ),
        ("from ib111 import week_12", ["config-file=ib111.toml"]),
        ("from ib111 import week_02  # noqa", ["config-file=ib111.toml"]),
        ("#from ib111 import week_02", ["config-file=ib111.toml"]),
    ],
)
def test_extract_args(mocker, contents, args):
    mock_contents(mocker, contents)
    assert extract_args("foo") == args


@pytest.fixture
def options() -> Dict[Option, OptionParse]:
    return {
        Option.SET_GROUPS: OptionParse(
            Option.SET_GROUPS, "", TakesVal.YES, [], Type.LIST, Combine.REPLACE
        ),
        Option.FLAKE8: OptionParse(Option.FLAKE8, "", TakesVal.YES, [], Type.STR, Combine.EXTEND),
        Option.CONFIG_FILE: OptionParse(
            Option.CONFIG_FILE, "", TakesVal.YES, DEFAULT_CONFIG, Type.STR, Combine.REPLACE
        ),
    }


@pytest.mark.parametrize(
    "raw,parsed",
    [
        (["set-groups=python-specific"], [UnprocessedArg(Option.SET_GROUPS, "python-specific")]),
        (["flake8=foo"], [UnprocessedArg(Option.FLAKE8, "foo")]),
        (["flake8="], [UnprocessedArg(Option.FLAKE8, "")]),
        (
            ["set-groups=python-specific", "flake8=foo"],
            [
                UnprocessedArg(Option.SET_GROUPS, "python-specific"),
                UnprocessedArg(Option.FLAKE8, "foo"),
            ],
        ),
        (["flake8=--enable=xxx"], [UnprocessedArg(Option.FLAKE8, "--enable=xxx")]),
        (["set-groups=ib111-week-02"], [UnprocessedArg(Option.SET_GROUPS, "ib111-week-02")]),
        (
            ["set-groups=ib111-week-12", "set-groups=ib111-week-02"],
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-12"),
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-02"),
            ],
        ),
        (["config=empty"], [UnprocessedArg(Option.CONFIG_FILE, "empty")]),
    ],
)
def test_parse_args(
    raw: List[str], options: Dict[Option, OptionParse], parsed: List[UnprocessedArg]
) -> None:
    assert parse_args(raw, options) == parsed


def packaged_config_files():
    packaged_config_dir = Path(__file__).parent / ".." / "edulint" / "config" / "files"
    return [c.stem for c in packaged_config_dir.iterdir() if c.suffix == ".toml"]


@pytest.mark.parametrize("config_name", packaged_config_files())
def test_packaged_configs_parse(config_name: str):
    assert parse_config_file(config_name, get_option_parses()) is not None


def test_remote_config_parses():
    url = remote_empty_config_url()
    assert parse_config_file(url, get_option_parses()) is not None


def relative_to_file(file_path, relative_path):
    dir_path = os.path.dirname(file_path)
    abspath = os.path.abspath(dir_path)
    return os.path.join(abspath, relative_path)

def test_local_config_found_relative_to_file():
    filename = get_tests_path(str(Path() / "file" / "somewhere" / "deep" / "03-d4_points.py"))
    iconfig, _lang_translations = get_config_one(filename, [])

    assert iconfig[Option.CONFIG_FILE] == relative_to_file(filename, "enable-missing-docstring.toml")
    assert iconfig[Option.PYLINT][-1] == "--enable=missing-function-docstring"


def test_local_configs_found_relative_to_files():
    filename1 = get_tests_path(str(Path() / "file" / "somewhere" / "deep" / "03-d4_points.py"))
    filename2 = get_tests_path(str(Path() / "file" / "somewhere" / "else" / "03-d4_points.py"))
    iconfigs = get_config_many([filename1, filename2], [])

    assert len(iconfigs) == 2
    iconfig1 = iconfigs[0][1]
    iconfig2 = iconfigs[1][1]

    assert iconfig1[Option.CONFIG_FILE] == relative_to_file(filename1, "enable-missing-docstring.toml")
    assert iconfig1[Option.PYLINT][-1] == "--enable=missing-function-docstring"

    assert iconfig2[Option.CONFIG_FILE] == relative_to_file(filename2, "enable-missing-docstring.toml")
    assert iconfig2[Option.PYLINT][-1] == "--enable=missing-module-docstring"


@pytest.fixture
def option_sets() -> OptionSets:
    return {
        "enhancement": OptionSet({Linter.PYLINT: ["aaa"]}),
        "python-specific": OptionSet({Linter.PYLINT: ["bbb"]}),
        Option.ALLOWED_ONECHAR_NAMES.to_name(): OptionSet({Linter.PYLINT: ["ccc"]}),
        "ib111-week-00": OptionSet({Linter.PYLINT: ["iii"]}),
        "ib111-week-02": OptionSet({Linter.PYLINT: ["kkk"]}),
        "ib111-week-12": OptionSet({Linter.PYLINT: ["lll"]}),
    }


@pytest.mark.parametrize(
    "args,config",
    [
        (
            [UnprocessedArg(Option.SET_GROUPS, "enhancement")],
            [
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
                ProcessedArg(Option.PYLINT, ["aaa"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "enhancement"),
                UnprocessedArg(Option.PYLINT, "zzz"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
                ProcessedArg(Option.PYLINT, ["aaa", "zzz"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.PYLINT, "zzz"),
                UnprocessedArg(Option.SET_GROUPS, "enhancement"),
            ],
            [
                ProcessedArg(Option.PYLINT, ["zzz", "aaa"]),
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.FLAKE8, "zzz"),
                UnprocessedArg(Option.SET_GROUPS, "enhancement"),
            ],
            [
                ProcessedArg(Option.FLAKE8, ["zzz"]),
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
                ProcessedArg(Option.PYLINT, ["aaa"]),
            ],
        ),
        (
            [UnprocessedArg(Option.ALLOWED_ONECHAR_NAMES, "n")],
            [ProcessedArg(Option.ALLOWED_ONECHAR_NAMES, "n"), ProcessedArg(Option.PYLINT, ["ccc"])],
        ),
        (
            [UnprocessedArg(Option.SET_GROUPS, "ib111-week-02")],
            [
                ProcessedArg(Option.SET_GROUPS, ["ib111-week-02"]),
                ProcessedArg(Option.PYLINT, ["kkk"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-12"),
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-02"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["ib111-week-02"]),
                ProcessedArg(Option.PYLINT, ["kkk"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-02"),
                UnprocessedArg(Option.PYLINT, "aaa"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["ib111-week-02"]),
                ProcessedArg(Option.PYLINT, ["kkk", "aaa"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-12"),
                UnprocessedArg(Option.SET_GROUPS, "enhancement"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
                ProcessedArg(Option.PYLINT, ["aaa"]),
            ],
        ),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-12,enhancement"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["ib111-week-12", "enhancement"]),
                ProcessedArg(Option.PYLINT, ["lll", "aaa"]),
            ],
        ),
        (
            [UnprocessedArg(Option.IGNORE_INFILE_CONFIG_FOR, "pylint,flake8")],
            [ProcessedArg(Option.IGNORE_INFILE_CONFIG_FOR, ["pylint", "flake8"])],
        ),
        (
            [UnprocessedArg(Option.IGNORE_INFILE_CONFIG_FOR, "")],
            [ProcessedArg(Option.IGNORE_INFILE_CONFIG_FOR, [])],
        ),
        ([UnprocessedArg(Option.SET_GROUPS, "")], [ProcessedArg(Option.SET_GROUPS, [])]),
        (
            [
                UnprocessedArg(Option.SET_GROUPS, "ib111-week-12"),
                UnprocessedArg(Option.SET_GROUPS, "enhancement"),
            ],
            [
                ProcessedArg(Option.SET_GROUPS, ["enhancement"]),
                ProcessedArg(Option.PYLINT, ["aaa"]),
            ],
        ),
    ],
)
def test_combine_and_translate_translates(
    args: List[UnprocessedArg], option_sets: OptionSets, config: List[ProcessedArg]
) -> None:
    def fill_in_defaults(
        config: List[ProcessedArg], option_parses: Dict[Option, OptionParse]
    ) -> List[ProcessedArg]:
        result = [ProcessedArg(o, Config._to_immutable(option_parses[o].default)) for o in Option]
        for arg in config:
            result[int(arg.option)] = ProcessedArg(arg.option, Config._to_immutable(arg.val))
        return tuple(result)

    option_parses = get_option_parses()
    result = Config("test", args, option_parses).to_immutable(option_sets)
    reference = fill_in_defaults(config, option_parses)
    assert result.config == reference


def _fill_in_file_config(config: Config) -> Config:
    config_file_result = parse_config_file(
        config.get_last_value(Option.CONFIG_FILE, use_default=True), get_option_parses()
    )
    assert config_file_result is not None
    file_config, option_sets, _lang_translations = config_file_result
    return Config.combine(file_config, config).to_immutable(option_sets)


def _arg_to_str(option: Option, val: Optional[str]) -> str:
    if val is None:
        return option.to_name()
    return f"{option.to_name()}={val}"


@pytest.mark.parametrize(
    "filename,cmd,args",
    [
        ("custom_set_empty_config.py", [], [Arg(Option.CONFIG_FILE, "empty")]),
        (
            "custom_set_empty_config.py",
            [_arg_to_str(Option.CONFIG_FILE, "default")],
            [Arg(Option.CONFIG_FILE, "default")],
        ),
        (
            "custom_set_empty_config.py",
            [_arg_to_str(Option.IGNORE_INFILE_CONFIG_FOR, Linter.EDULINT.to_name())],
            [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "edulint"), Arg(Option.CONFIG_FILE, "default")],
        ),
        (
            "custom_set_empty_config.py",
            [_arg_to_str(Option.IGNORE_INFILE_CONFIG_FOR, "all")],
            [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "all"), Arg(Option.CONFIG_FILE, "default")],
        ),
        ("custom_set_ignore_infile_and_some.py", [], [Arg(Option.IGNORE_INFILE_CONFIG_FOR, "all")]),
        ("custom_set_replace_option.py", [], [Arg(Option.SET_GROUPS, "ib111-week-01")]),
        (
            "custom_set_replace_option.py",
            [_arg_to_str(Option.SET_GROUPS, "ib111-week-05")],
            [
                Arg(Option.SET_GROUPS, "ib111-week-01"),
                Arg(Option.SET_GROUPS, "ib111-week-05"),
            ],
        ),
    ],
)
def test_get_config_one(filename: str, cmd: List[str], args: List[UnprocessedArg]):
    iconfig = _fill_in_file_config(Config("test", args))
    assert get_config_one(get_tests_path(filename), cmd)[0].config == iconfig.config


@pytest.mark.parametrize(
    "filenames,partition",
    [
        (
            [
                Path("tests/data/custom_nonpep_assign.py"),
                Path("tests/data/custom_flake8_pylint_config.py"),
            ],
            [
                ([Path("tests/data/custom_nonpep_assign.py")], Config(enabler=None)),
                (
                    [Path("tests/data/custom_flake8_pylint_config.py")],
                    Config(
                        enabler=None,
                        config=[Arg(Option.PYLINT, "--enable=missing-module-docstring")],
                    ),
                ),
            ],
        )
    ],
)
def test_get_config_many(filenames: List[str], partition: List[Tuple[List[str], Config]]):
    configs = get_config_many(filenames, [])
    assert len(configs) == len(partition)

    for i in range(len(configs)):
        fns1, iconfig1, _lang_translations = configs[i]
        fns2, config2 = partition[i]

        assert fns1 == fns2

        iconfig2 = _fill_in_file_config(config2)
        assert iconfig1.config == iconfig2.config



from edulint.options import Option
from edulint.config.arg import UnprocessedArg
from edulint.config.config import Config, get_config_one
from edulint.linting.problem import Problem
from edulint.linting.linting import lint_one
from dataclasses import fields, replace
import os
import pathlib
from typing import List
import tempfile

LAZY_INT = -1
FILLER_CODE = ""


def lazy_equals(received: Problem, expected: Problem) -> None:
    if not any(expected.has_value(f.name) for f in fields(Problem)):
        assert False, f"unexpected problem {repr(received)}"

    copy = replace(received)
    for field in fields(Problem):
        if not expected.has_value(field.name):
            setattr(copy, field.name, getattr(expected, field.name))

    assert copy == expected


def lazy_problem() -> Problem:
    return Problem(None, "", LAZY_INT, LAZY_INT, None, "", LAZY_INT, LAZY_INT)  # type: ignore


def filler_problem() -> Problem:
    lazy = lazy_problem()
    lazy.code = FILLER_CODE
    return lazy


def fill(lst, len_):
    return lst + [filler_problem() for _ in range(len_ - len(lst))]


def lazy_equal(received: List[Problem], expected: List[Problem]) -> None:
    len_ = max(len(received), len(expected))
    for r, e in zip(fill(received, len_), fill(expected, len_)):
        lazy_equals(r, e)


def get_tests_path(filename: str) -> str:
    return str((pathlib.Path(__file__).parent / "data" / filename).resolve())


def remote_empty_config_url() -> str:
    return "https://raw.githubusercontent.com/GiraffeReversed/edulint/v2.9.2/edulint/config/files/empty.toml"


def prepare_config(filename: str, args: List[UnprocessedArg], from_empty: bool) -> Config:
    config_args = Config("test", args)
    args_config = config_args.get_last_value(Option.CONFIG_FILE, use_default=False)
    config_path = (
        args_config
        if args_config is not None
        else "empty"
        if from_empty
        else config_args.get_last_value(Option.CONFIG_FILE, use_default=True)
    )
    config = get_config_one(
        filename,
        [f"{arg.option.to_name()}={arg.val}" if arg.val is not None else arg.option.to_name() for arg in args]
        + [f"config-file={config_path}"]
    )
    assert config is not None
    return config


def just_lint(filename: str, args: List[UnprocessedArg], from_empty: bool = True) -> List[Problem]:
    path = get_tests_path(filename)
    config = prepare_config(path, args, from_empty)
    return lint_one(path, config)


def apply_and_lint(
    filename: str,
    args: List[UnprocessedArg],
    expected_output: List[Problem],
    from_empty: bool = True,
) -> None:
    lazy_equal(just_lint(filename, args, from_empty), expected_output)


def create_apply_and_lint(
    lines: List[str],
    args: List[UnprocessedArg],
    expected_output: List[Problem],
    from_empty: bool = True,
) -> None:
    tf = tempfile.NamedTemporaryFile("w+", delete=False, suffix=".py")
    try:
        tf.writelines([line + "\n" for line in lines])
        tf.close()
        apply_and_lint(tf.name, args, expected_output, from_empty)
    finally:
        os.remove(tf.name)



from pathlib import Path
import string
import os
from typing import Dict, Any, Optional
import hashlib
import json
import time
from loguru import logger
from enum import Enum, auto

import tomli
import requests
from platformdirs import PlatformDirs


ALLOWED_FILENAME_LETTERS = string.ascii_letters + string.digits + "-_"
ALLOW_UNRESTRICTED_LOCAL_PATHS = True
ALLOW_HTTP_S_PATHS = True


class EduLintConfigFileException(Exception):
    pass


class ConfigFileAccessMethodNotAllowedException(EduLintConfigFileException):
    pass


class InvalidConfigFile(EduLintConfigFileException):
    pass


class ConfigFileType(Enum):
    REMOTE = auto()
    LOCAL = auto()
    PACKAGED = auto()


def load_toml_file(filename_or_url: str) -> Optional[Dict[str, Any]]:
    try:
        file_content = _load_file_from_uri(filename_or_url)
    except Exception as e:
        logger.error("error locating config file '{f}':\n {e}", f=filename_or_url, e=e)
        return None

    try:
        file_toml = tomli.loads(file_content)
        return file_toml
    except tomli.TOMLDecodeError as e:
        logger.error("invalid TOML config file '{f}':\n {e}", f=filename_or_url, e=e)
        return None


def get_config_type(path_or_url: str) -> ConfigFileType:
    if path_or_url.startswith(("http://", "https://")):
        return ConfigFileType.REMOTE

    if _only_acceptable_chars(path_or_url):
        return ConfigFileType.PACKAGED

    return ConfigFileType.LOCAL


def get_path_relative_to(filepath: str, parent_filepath: str):
    config_type = get_config_type(filepath)
    if config_type != ConfigFileType.LOCAL or os.path.isabs(filepath):
        return filepath

    assert not os.path.isdir(parent_filepath)
    return os.path.abspath(os.path.join(os.path.dirname(parent_filepath), filepath))


def _load_file_from_uri(path_or_url: str) -> str:
    config_type = get_config_type(path_or_url)

    if config_type == ConfigFileType.REMOTE:
        return _load_external_config_file(path_or_url)

    if config_type == ConfigFileType.LOCAL:
        return _load_local_config_file(path_or_url, path_or_url, "found locally")

    if config_type == ConfigFileType.PACKAGED:
        return _load_packaged_config_file(path_or_url)

    assert False, "unreachable"


def _only_acceptable_chars(filepath: str) -> bool:
    return all([x in ALLOWED_FILENAME_LETTERS for x in filepath])


def _load_packaged_config_file(filename: str) -> str:
    assert _only_acceptable_chars(filename)

    relative_path = os.path.join(os.path.dirname(__file__), "files", filename + ".toml")
    return _load_local_config_file(relative_path, filename, "packaged", is_path_safe=True)


def _load_local_config_file(
    filepath: str, passed_name: str, message: str, is_path_safe: bool = False
) -> str:
    if not is_path_safe and not ALLOW_UNRESTRICTED_LOCAL_PATHS:
        raise ConfigFileAccessMethodNotAllowedException(
            "arbitrary local filepaths are not enabled."
        )

    # Doing the test in two steps should prevent possible exception during the if test.
    if not (Path(filepath).exists() and Path(filepath).is_file()):
        raise FileNotFoundError(f"configuration file '{passed_name}' not {message}.")

    with open(filepath, encoding="utf8") as f:
        return f.read()


def _load_external_config_file(url: str) -> str:
    if not ALLOW_HTTP_S_PATHS:
        raise ConfigFileAccessMethodNotAllowedException(
            "loading of external configs using HTTP/HTTPS is disallowed in EduLint's configuration."
        )

    # can throw exception FileNotFoundError if remote URL didn't work and file is not cached yet
    return CachedHTTPGet.http_get(url)


class CachedHTTPGet:
    @classmethod
    def http_get(
        cls,
        url: str,
        max_cache_time: int = 5 * 60,
        max_cache_time_when_offline: int = 500 * 24 * 60,
    ) -> str:
        """
        Source priority: file cache with max age > HTTP GET from URL > file cache with extended max age
        """

        cached_version: str = cls._read_version_from_disk(url, max_age_in_seconds=max_cache_time)
        if cached_version:
            return cached_version

        try:
            resp = requests.get(url, timeout=5)
            if resp.status_code == 200:
                content = resp.text
                cls._write_version_to_disk(url, content)
                return content

            logger.error(
                "request for external config '{url}' failed with status code {code}. "
                "Trying to fallback to cached version if available.",
                url=url,
                code=resp.status_code,
            )

        except requests.exceptions.RequestException:
            pass

        cached_version: str = cls._read_version_from_disk(
            url, max_age_in_seconds=max_cache_time_when_offline
        )
        if cached_version:
            return cached_version
        raise FileNotFoundError(
            f"request for external config '{url}' failed -- maybe you are offline or the URL is incorrect."
        )

    @staticmethod
    def _get_timestamp() -> int:
        return int(time.time())

    @staticmethod
    def _get_edulint_cache_folder_location() -> Path:
        path_str = PlatformDirs(appname="edulint").user_data_dir
        path = Path(path_str)
        path.mkdir(parents=True, exist_ok=True)
        return path

    @staticmethod
    def _convert_source_to_safe_filename(source: str) -> str:
        return hashlib.sha256(source.encode(errors="replace")).hexdigest() + ".toml"

    @classmethod
    def _get_filepath_from_source(cls, source: str) -> str:
        folder_path = cls._get_edulint_cache_folder_location()
        source_filename = cls._convert_source_to_safe_filename(source)
        # future: saving everything into same folder can become problem if there are too many filder in the folder
        return folder_path / source_filename

    @classmethod
    def _get_metadata_filepath(cls, source: str) -> str:
        base_filepath = cls._get_filepath_from_source(source)
        return base_filepath.with_suffix(base_filepath.suffix + ".metadata")

    @classmethod
    def _write_version_to_disk(cls, source: str, content: str):
        try:
            with open(cls._get_filepath_from_source(source), "w", encoding="utf8") as f:
                # security: writing arbitrary content from web is not great, but at least it's written as text
                f.write(content)
            with open(cls._get_metadata_filepath(source), "w", encoding="utf8") as f:
                json.dump(
                    {
                        "timestamp": cls._get_timestamp(),
                        "source": source,
                    },
                    f,
                    indent=4,
                )
        except Exception as e:
            logger.error("saving cache file for configuration failed:\n {e}", e=e)

    @classmethod
    def _read_version_from_disk(
        cls, source: str, max_age_in_seconds: int = 5 * 60
    ) -> Optional[str]:
        try:
            with open(cls._get_metadata_filepath(source), "r", encoding="utf8") as f:
                metadata = json.load(f)
            if metadata["timestamp"] + max_age_in_seconds <= cls._get_timestamp():  # todo: int
                return None

            with open(cls._get_filepath_from_source(source), "r", encoding="utf8") as f:
                return f.read()
        except FileNotFoundError:
            return None
        except Exception as e:
            logger.error("reading or parsing cache file for configuration failed:\n {e}", e=e)

        return None



from edulint.linters import Linter
from dataclasses import dataclass, field, fields
from dataclasses_json import dataclass_json, config
from marshmallow import fields as mm_fields
from typing import Optional, Dict, Union, List

ProblemJson = Dict[str, Union[str, int]]


@dataclass_json
@dataclass
class Problem:
    source: Linter = field(
        metadata=config(encoder=Linter.to_name, decoder=Linter.from_name, mm_field=mm_fields.Str())
    )
    enabled_by: Optional[str]
    path: str
    line: int
    column: int
    code: str
    text: str
    end_line: Optional[int] = None
    end_column: Optional[int] = None
    symbol: Optional[str] = None

    def set_source(self, v: Linter) -> "Problem":
        self.source = v
        return self

    def set_enabled_by(self, v: Optional[str]) -> "Problem":
        self.enabled_by = v
        return self

    def set_path(self, v: str) -> "Problem":
        self.path = v
        return self

    def set_line(self, v: int) -> "Problem":
        self.line = v
        return self

    def set_column(self, v: int) -> "Problem":
        self.column = v
        return self

    def set_code(self, v: str) -> "Problem":
        self.code = v
        return self

    def set_symbol(self, v: str) -> "Problem":
        self.symbol = v
        return self

    def set_text(self, v: str) -> "Problem":
        self.text = v
        return self

    def set_end_line(self, v: Optional[int]) -> "Problem":
        self.end_line = v
        return self

    def set_end_column(self, v: Optional[int]) -> "Problem":
        self.end_column = v
        return self

    def has_value(self, attr: str) -> bool:
        val = getattr(self, attr)
        return (
            val is not None
            and (not isinstance(val, str) or bool(val))
            and (not isinstance(val, int) or val >= 0)
        )

    def __repr__(self) -> str:
        def get_attrvals() -> List[str]:
            result = []
            for f in fields(Problem):
                attr = f.name
                if self.has_value(attr):
                    attrval = getattr(self, attr)
                    attrval = attrval if not isinstance(attrval, str) else f'"{attrval}"'
                    result.append(f"{attr}={attrval}")
            return result

        return f"Problem({', '.join(get_attrvals())})"

    def __str__(self) -> str:
        enabler_str = f" [{self.enabled_by}]" if self.enabled_by is not None else ""
        return f"{self.path}:{self.line}:{self.column}: {self.code} {self.text}{enabler_str}"



from typing import Union, Tuple, List, Any, Iterator, Callable, Optional
import copy
from enum import Enum

from astroid import nodes

from edulint.linting.analyses.cfg.utils import syntactic_children_locs_from, get_cfg_loc
from edulint.linting.analyses.data_dependency import (
    Variable,
    get_vars_defined_before,
    get_vars_used_after,
    MODIFYING_EVENTS,
)
from edulint.linting.analyses.cfg.visitor import CFGVisitor
from edulint.linting.checkers.utils import new_node, eprint


class AunifyVar(nodes.Name):
    def __init__(self, name: str):
        self.name = name.upper()
        self.parent = None
        self.lineno = 0
        self.subs = []
        self.sub_locs = []

    def __repr__(self):
        return self.name

    def accept(self, visitor):
        return getattr(visitor, "visit_name")(self)

    def isdigit(self):
        return False

    def __len__(self):
        return 1

    def __str__(self):
        return self.name

    def __add__(self, other):
        return str(self) + other

    def __radd__(other, self):
        return other + str(self)

    def replace(self, old, new):
        return self.name.replace(old, new)


def _to_list(val):
    return val if isinstance(val, list) else [val]


class DisallowedAntiunification(Exception):
    pass


ASTROID_EXTRA_FIELDS = {
    nodes.Name: ("name",),
    nodes.AssignName: ("name",),
    nodes.Attribute: ("attrname",),
    nodes.AssignAttr: ("attrname",),
    nodes.BoolOp: ("op",),
    nodes.BinOp: ("op",),
    nodes.UnaryOp: ("op",),
    nodes.AugAssign: ("op",),
    nodes.FunctionDef: ("name",),
    nodes.Const: ("value",),
    nodes.Nonlocal: ("names",),
    nodes.Global: ("names",),
    nodes.ImportFrom: (
        "modname",
        "names",
    ),
    nodes.Import: ("names",),
    nodes.Keyword: ("arg",),
    nodes.Subscript: ("ctx",),
    nodes.Starred: ("ctx",),
    nodes.DelName: ("name",),
    nodes.DelAttr: ("attrname",),
}


def get_all_fields(node):
    return type(node)._astroid_fields + ASTROID_EXTRA_FIELDS.get(type(node), ())


class Antiunify:
    __num = 0

    def _get_new_avar(self, extra: str = None):
        self.__num += 1
        return AunifyVar(f"id_{self.__num}{('_' + extra) if extra is not None else ''}")

    def _new_aunifier(
        self, to_aunify: List[Any], stop_on: Callable[[List[AunifyVar]], bool], extra: str = None
    ):
        avar = self._get_new_avar(extra)
        avar.subs = to_aunify.copy()
        avar.sub_locs = [n.cfg_loc for n in to_aunify if hasattr(n, "cfg_loc")]

        if stop_on([avar]):
            raise DisallowedAntiunification()

        return avar, [avar]

    def _aunify_consts(self, vals: List[Any], stop_on: Callable[[List[AunifyVar]], bool]):
        if all(v == vals[0] for v in vals):
            return vals[0], []

        return self._new_aunifier(vals, stop_on)

    def _aunify_avars(self, to_aunify: List[Any], stop_on: Callable[[List[AunifyVar]], bool]):
        core = self._get_new_avar()

        for node in to_aunify:
            if isinstance(node, AunifyVar):
                core.subs.extend(node.subs)
                core.sub_locs.extend(node.sub_locs)
            else:
                core.subs.append(node)
                if hasattr(node, "cfg_loc"):
                    core.sub_locs.append(node.cfg_loc)

        if stop_on([core]):
            raise DisallowedAntiunification()

        return core, [core]

    def antiunify(
        self, to_aunify: List[Any], stop_on: Callable[[List[AunifyVar]], bool]
    ) -> Optional[Tuple[Any, List[AunifyVar]]]:
        if any(isinstance(n, AunifyVar) for n in to_aunify):
            return self._aunify_avars(to_aunify, stop_on)

        if isinstance(to_aunify[0], (list, tuple)):
            assert all(isinstance(n, (list, tuple)) for n in to_aunify)
            return self._antiunify_lists(to_aunify, stop_on)

        if not any(isinstance(n, nodes.NodeNG) for n in to_aunify):
            return self._aunify_consts(to_aunify, stop_on)

        if not all(isinstance(n, type(to_aunify[0])) for n in to_aunify):
            return self._new_aunifier(
                to_aunify, stop_on, extra="-".join(type(n).__name__ for n in to_aunify)
            )

        # astroid.nodes of same type
        some = to_aunify[0]
        all_fields = get_all_fields(some)

        return self._aunify_by_attrs(to_aunify, all_fields, stop_on)

    def _antiunify_lists(
        self,
        to_aunify: List[List[Any]],
        stop_on: Callable[[List[AunifyVar]], bool],
        attr: str = "<none>",
    ):
        if not all(len(n) == len(to_aunify[0]) for n in to_aunify):
            attr_core, attr_avars = self._new_aunifier(
                to_aunify,
                stop_on,
                extra=f"{attr}-" + "-".join(str(len(n)) for n in to_aunify),
            )
            some = [v for n in to_aunify for v in n][0]
            if isinstance(some, tuple):
                fst, snd = some
                if isinstance(fst, str) and isinstance(snd, (str, type(None))):
                    return [(attr_core, None)], attr_avars
                elif isinstance(fst, str) and isinstance(snd, nodes.NodeNG):
                    return [("", attr_core)], attr_avars
                elif isinstance(fst, nodes.NodeNG) and isinstance(snd, nodes.NodeNG):
                    other_core, _ = self._new_aunifier(
                        [], stop_on, extra=f"{attr}-" + "-".join(str(len(n)) for n in to_aunify)
                    )
                    return [(attr_core, other_core)], attr_avars
            return [attr_core], attr_avars

        core = []
        avars = []
        for i in range(len(to_aunify[0])):
            children = [n[i] for n in to_aunify]

            child_core, child_avars = self.antiunify(children, stop_on)
            child_core = tuple(child_core) if isinstance(children[0], tuple) else child_core

            core.append(child_core)
            avars.extend(child_avars)

        return core, avars

    def _aunify_many_attrs(
        self, attrs: List[str], to_aunify: List[Any], stop_on: Callable[[List[AunifyVar]], bool]
    ):
        avars = []
        attr_cores = {}

        for attr in attrs:
            attr_vals = [getattr(n, attr) for n in to_aunify]

            attr_core, attr_avars = self.antiunify(attr_vals, stop_on)
            attr_cores[attr] = attr_core

            avars.extend(attr_avars)

        return attr_cores, avars

    def _aunify_by_attrs(
        self,
        to_aunify,
        attrs: List[str],
        stop_on: Callable[[List[AunifyVar]], bool],
    ):
        some = to_aunify[0]
        assert all(isinstance(n, type(some)) for n in to_aunify)

        attr_cores, avars = self._aunify_many_attrs(attrs, to_aunify, stop_on)

        core = new_node(type(some), **attr_cores)

        # pylint overloads __getitem__ on constants, so hasattr would fail
        try:
            assert all(hasattr(n, "cfg_loc") or hasattr(n, "sub_locs") for n in to_aunify)
            core.sub_locs = [
                loc
                for node in to_aunify
                for loc in (
                    [getattr(node, "cfg_loc")]
                    if hasattr(node, "cfg_loc")
                    else getattr(node, "sub_locs")
                )
            ]
        except AssertionError:
            pass

        for attr_core in attr_cores.values():
            set_parents(core, attr_core, recursive=False)

        if stop_on(avars):
            raise DisallowedAntiunification()

        return core, avars


def contains_avar(node: Union[nodes.NodeNG, List[nodes.NodeNG]], avars):
    if isinstance(node, nodes.NodeNG):
        ns = [node]
    elif not isinstance(node, (list, tuple)) or len(node) == 0:
        return False
    elif not isinstance(node[0], nodes.NodeNG):
        ns = [n for subn in node for n in subn]
    else:
        ns = node

    for avar in avars:
        for ancestor in avar.node_ancestors():
            if any(ancestor == n for n in ns):
                return True
    return False


def get_avars(core: nodes.NodeNG, result: List[AunifyVar] = None) -> List[AunifyVar]:
    result = result if result is not None else []

    if isinstance(core, AunifyVar):
        result.append(core)

    elif isinstance(core, nodes.NodeNG):
        all_fields = get_all_fields(core)
        for attr in all_fields:
            get_avars(getattr(core, attr), result)

    elif isinstance(core, (tuple, list)):
        for n in core:
            get_avars(n, result)

    return result


def ancestors_till(node: nodes.NodeNG, ancestor: nodes.NodeNG):
    result = []
    for parent in node.node_ancestors():
        if parent == ancestor:
            break
        result.append(parent)
    return result


def get_avar_parent_in_sub(avar_node, sub_node, avar):
    if avar_node == avar.parent:
        return sub_node

    if isinstance(avar_node, list):
        assert isinstance(sub_node, list) and len(avar_node) == len(sub_node)
        for an, sn in zip(avar_node, sub_node):
            if contains_avar(an, [avar]):
                return get_avar_parent_in_sub(an, sn, avar)
    else:
        for attr in get_all_fields(avar_node):
            avar_attr_node = getattr(avar_node, attr)
            if contains_avar(avar_attr_node, [avar]):
                return get_avar_parent_in_sub(avar_attr_node, getattr(sub_node, attr), avar)

    assert False, f"unreachable, but {avar}"


def sub_to_variable(avar, i):
    assert isinstance(avar.parent, nodes.AssignName)

    sub = avar.subs[i]
    avar_loc = get_cfg_loc(avar)
    sub_loc = avar_loc.node.sub_locs[i]

    event_nodes = list(
        {
            (var.scope, event.node)
            for var, event in sub_loc.var_events.for_name(sub)
            if event.type in MODIFYING_EVENTS
        }
    )
    if len(event_nodes) == 1:
        return Variable(sub, event_nodes[0][0])
    assert len(event_nodes) > 1

    sub_node = get_avar_parent_in_sub(avar_loc.node, sub_loc.node, avar)

    for scope, node in event_nodes:
        if node == sub_node:
            return Variable(sub, scope)

    assert False, f"unreachable, but {sub}"


def get_removable(core, avar) -> bool:
    assert isinstance(avar.parent, nodes.AssignName)
    avar_loc = get_cfg_loc(avar)

    to_remove = {avar}

    for ic, core_loc in enumerate(syntactic_children_locs_from(avar_loc, core)):
        if ic == 0:
            continue
        for il, loc in enumerate(core_loc.node.sub_locs):
            var = sub_to_variable(avar, il)
            if var in loc.var_events:
                for loc_avar in get_avars(core_loc.node):
                    assert len(avar.subs) == len(loc_avar.subs)
                    if any(asub != lasub for asub, lasub in zip(avar.subs, loc_avar.subs)):
                        return set()
                    to_remove.add(loc_avar)
    return to_remove


def remove_renamed_identical_vars(core, avars: List[AunifyVar]):
    if not any(isinstance(avar.parent, nodes.AssignName) for avar in avars):
        return core, avars

    vars_defined_before = get_vars_defined_before(core)
    vars_used_after = get_vars_used_after(core)
    irremovable_vars = set(vars_defined_before.keys()) | set(vars_used_after.keys())

    for avar in avars:
        if not isinstance(avar.parent, nodes.AssignName) or any(
            sub_to_variable(avar, i) in irremovable_vars for i, sub in enumerate(avar.subs)
        ):
            continue

        for to_remove_avar in get_removable(core, avar):
            to_remove_avar.parent.name = avar.subs[0]

    return core, get_avars(core)


def antiunify(
    to_aunify: List[Union[nodes.NodeNG, List[nodes.NodeNG]]],
    stop_on: Callable[[List[AunifyVar]], bool] = lambda _: False,
    stop_on_after_renamed_identical: Callable[[List[AunifyVar]], bool] = lambda _: False,
) -> Optional[Tuple[Any, List[AunifyVar]]]:
    try:
        core, avars = Antiunify().antiunify(to_aunify, stop_on)
    except DisallowedAntiunification:
        return None

    wrapper = new_node(nodes.Module, name="tmp", body=core if isinstance(core, list) else [core])
    wrapper.accept(CFGVisitor())

    core, avars = remove_renamed_identical_vars(core, avars)
    if stop_on_after_renamed_identical(avars):
        return None
    return core, avars


def set_parents(parent: nodes.NodeNG, node: Any, recursive):
    if isinstance(node, nodes.NodeNG):
        node.parent = parent

        if recursive:
            for attr in get_all_fields(node):
                children = getattr(node, attr)
                set_parents(node, children, recursive)

    elif isinstance(node, (list, tuple)):
        for elem in node:
            set_parents(parent, elem, recursive)

    elif node is not None and not isinstance(
        node, (str, bool, int, float, bytes, type(Ellipsis), Enum)
    ):
        assert False, f"unreachable, but {type(node)}"


def get_sub_variant(core, index: int):
    if isinstance(core, (list, tuple)):
        return type(core)([get_sub_variant(c, index) for c in core])

    if not isinstance(core, nodes.NodeNG):
        return core

    if isinstance(core, AunifyVar):
        variant = core.subs[index]

    else:
        attr_variants = {
            attr: get_sub_variant(getattr(core, attr), index) for attr in get_all_fields(core)
        }
        variant = new_node(type(core), **attr_variants)

        for attr, attr_variant in attr_variants.items():
            set_parents(variant, attr_variant, recursive=False)

    try:
        if len(core.sub_locs) > 0:
            variant.cfg_loc = core.sub_locs[index]
    except AttributeError:
        pass

    return variant


##############################################################################
####                            to string                                 ####
##############################################################################


class AunifyVarAsString(nodes.as_string.AsStringVisitor):
    @staticmethod
    def from_aunifyvar(val):
        if isinstance(val, str):
            return val
        return val.name

    def visit_assign(self, node) -> str:
        """return an astroid.Assign node as string"""
        lhs = " = ".join(AunifyVarAsString.from_aunifyvar(n.accept(self)) for n in node.targets)
        return f"{lhs} = {node.value.accept(self)}"

    def visit_call(self, node) -> str:
        """return an astroid.Call node as string"""
        return super().visit_call(node)

    def visit_name(self, node) -> str:
        if isinstance(node.name, str):
            return node.name
        return node.name.name

    def visit_assignname(self, node) -> str:
        if isinstance(node.name, str):
            return node.name
        return node.name.name

    def _should_wrap(self, node, child, is_left: bool) -> bool:
        try:
            return super()._should_wrap(node, child, is_left)
        except KeyError:
            return True

    def visit_tuple(self, node) -> str:
        """return an astroid.Tuple node as string"""
        if len(node.elts) == 1:
            return f"({node.elts[0].accept(self)}, )"
        return f"({', '.join(child.accept(self) for child in node.elts)})"

    def visit_arguments(self, node) -> str:
        node_copy = copy.copy(node)
        node_copy.args = []
        for arg in node.args:
            if isinstance(arg, nodes.AssignName) and not isinstance(arg.name, str):
                arg_copy = copy.copy(arg)
                arg_copy.name = str(arg.name)
                node_copy.args.append(arg_copy)
            else:
                node_copy.args.append(arg)
        return node_copy.format_args()

    def visit_compare(self, node) -> str:
        """return an astroid.Compare node as string"""
        if not isinstance(node.ops, list):
            return f"{self._precedence_parens(node, node.left)} {node.ops}"

        rhs = []
        for pair in node.ops:
            if isinstance(pair, tuple):
                op, expr = pair
                rhs.append(f"{op} {self._precedence_parens(node, expr, is_left=False)}")
            else:
                rhs.append(pair.accept(self))
        return f"{self._precedence_parens(node, node.left)} {' '.join(rhs)}"

    def visit_dict(self, node) -> str:
        """return an astroid.Dict node as string"""
        return "{%s}" % ", ".join(self._visit_dict(node))

    def _visit_dict(self, node) -> Iterator[str]:
        for pair in node.items:
            if isinstance(pair, tuple):
                key, value = pair
                key = key.accept(self)
                value = value.accept(self)
                if key == "**":
                    # It can only be a DictUnpack node.
                    yield key + value
                else:
                    yield f"{key}: {value}"
            else:
                yield pair.name

    def visit_importfrom(self, node) -> str:
        def _import_string(names) -> str:
            """return a list of (name, asname) formatted as a string"""
            _names = []
            for name, asname in names:
                if asname is not None:
                    _names.append(f"{name} as {asname}")
                else:
                    _names.append(name)
            return ", ".join(n if isinstance(n, str) else str(n) for n in _names)

        """return an astroid.ImportFrom node as string"""
        return "from {} import {}".format(
            "." * (node.level or 0) + node.modname, _import_string(node.names)
        )


def core_as_string(n) -> str:
    visitor = AunifyVarAsString()
    if isinstance(n, list):
        return "\n".join(n.accept(visitor) for n in n)
    return n.accept(visitor)


def cprint(n) -> None:
    eprint(core_as_string(n))


def eprint_aunify_core(n, depth: int = 0):
    def get_name(n):
        if isinstance(n, nodes.Const):
            return None
        if hasattr(n, "name"):
            return getattr(n, "name")
        if hasattr(n, "attrname"):
            return getattr(n, "attrname")
        return None

    if n is None:
        return

    eprint(depth * " ", type(n).__name__, sep="", end="")

    name = get_name(n)
    if name is not None:
        eprint(".", name, sep="", end="")
    eprint()

    for attr in n._astroid_fields:
        eprint((depth + 1) * " ", attr, sep="")
        children = getattr(n, attr)
        if not isinstance(children, list):
            children = [children]

        for child in children:
            if isinstance(child, tuple):
                op, child = child
                eprint((depth + 2) * " ", op)
            eprint_aunify_core(child, depth + 2)
